# -*- coding: utf-8 -*-
"""The MongoDB vector store implementation using MongoDB Vector Search.

This implementation provides a vector database store using MongoDB's vector
 search capabilities. It requires MongoDB with vector search support and
 automatically creates vector search indexes.
"""
import asyncio
import time

from typing import Any, Literal, TYPE_CHECKING

from .._reader import Document
from ._store_base import VDBStoreBase
from .._document import DocMetadata
from ...types import Embedding

if TYPE_CHECKING:
    from pymongo import AsyncMongoClient
else:
    AsyncMongoClient = "pymongo.AsyncMongoClient"


class MongoDBStore(VDBStoreBase):
    """MongoDB vector store using MongoDB Vector Search.

    This class provides a vector database store implementation using MongoDB's
    vector search capabilities. It requires MongoDB with vector search support
    and creates vector search indexes automatically.

    .. note:: Ensure your MongoDB instance supports Vector Search
    functionality.

    .. note:: The store automatically creates database, collection, and vector
    search index on first operation. No manual initialization is required.
    """

    def __init__(
        self,
        host: str,
        db_name: str,
        collection_name: str,
        dimensions: int,
        index_name: str = "vector_index",
        distance: Literal["cosine", "euclidean", "dotProduct"] = "cosine",
        filter_fields: list[str] | None = None,
        client_kwargs: dict[str, Any] | None = None,
        db_kwargs: dict[str, Any] | None = None,
        collection_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Initialize the MongoDB vector store.

        Args:
            host (`str`):
                MongoDB connection host, e.g., "mongodb://localhost:27017" or
                "mongodb+srv://cluster.mongodb.net/".
            db_name (`str`):
                Database name to store vector documents.
            collection_name (`str`):
                Collection name to store vector documents.
            dimensions (`int`):
                Embedding dimensions for the vector search index.
            index_name (`str`, defaults to "vector_index"):
                Vector search index name.
            distance (`Literal["cosine", "euclidean", "dotProduct"]`, \
            defaults to "cosine"):
                Distance metric for vector similarity. Can be one of "cosine",
                "euclidean", or "dotProduct".
            filter_fields (`list[str] | None`, optional):
                List of field paths to index for filtering in $vectorSearch.
                For example: ["payload.doc_id", "payload.chunk_id"].
                These fields can then be used in the `filter` parameter of
                the `search` method. MongoDB $vectorSearch filter supports:
                $gt, $gte, $lt, $lte, $eq, $ne, $in, $nin, $exists, $not.
            client_kwargs (`dict[str, Any] | None`, optional):
                Additional kwargs for MongoDB client.
            db_kwargs (`dict[str, Any] | None`, optional):
                Additional kwargs for database.
            collection_kwargs (`dict[str, Any] | None`, optional):
                Additional kwargs for collection.

        Raises:
            ImportError: If pymongo is not installed.
        """
        try:
            from pymongo import AsyncMongoClient
        except ImportError as e:
            raise ImportError(
                "Please install the latest pymongo package to use "
                "AsyncMongoClient: `pip install pymongo`",
            ) from e

        self._client: AsyncMongoClient = AsyncMongoClient(
            host,
            **(client_kwargs or {}),
        )
        self.db_name = db_name
        self.collection_name = collection_name
        self.index_name = index_name
        self.dimensions = dimensions
        self.distance = distance
        self.filter_fields = filter_fields or []
        self.db_kwargs = db_kwargs or {}
        self.collection_kwargs = collection_kwargs or {}

        self._db = None
        self._collection = None

    async def _validate_db_and_collection(self) -> None:
        """Validate the database and collection exist, create if necessary.

        This method ensures the database and collection are available,
        and creates a vector search index for the collection.

        Raises:
            Exception: If database or collection creation fails.
        """
        self._db = self._client.get_database(
            self.db_name,
            **self.db_kwargs,
        )

        if self.collection_name not in await self._db.list_collection_names():
            self._collection = await self._db.create_collection(
                self.collection_name,
            )
        else:
            self._collection = self._db.get_collection(
                self.collection_name,
                **self.collection_kwargs,
            )

        from pymongo.operations import SearchIndexModel

        # Build index fields: vector field + optional filter fields
        index_fields: list[dict[str, Any]] = [
            {
                "type": "vector",
                "path": "vector",
                "similarity": self.distance,
                "numDimensions": self.dimensions,
            },
        ]

        # Add user-specified filter fields
        for field_path in self.filter_fields:
            index_fields.append(
                {
                    "type": "filter",
                    "path": field_path,
                },
            )

        search_index_model = SearchIndexModel(
            definition={"fields": index_fields},
            name=self.index_name,
            type="vectorSearch",
        )

        await self._collection.create_search_index(
            model=search_index_model,
        )

    async def _wait_for_index_ready(self, timeout: int = 30) -> None:
        """Wait for the vector search index to be ready with timeout
        protection.

        Args:
            timeout (`int`, defaults to 30):
                Maximum time to wait in seconds.

        Raises:
            TimeoutError: If index is not ready within the timeout period.
        """

        start_time = time.time()
        while time.time() - start_time < timeout:
            try:
                indices = []
                async for idx in await self._collection.list_search_indexes(
                    self.index_name,
                ):
                    indices.append(idx)
                if indices and indices[0].get("queryable") is True:
                    return
            except Exception:
                pass

            await asyncio.sleep(0.2)

        raise TimeoutError(f"Index not ready after {timeout} seconds")

    async def add(self, documents: list[Document], **kwargs: Any) -> None:
        """Insert documents with embeddings into MongoDB.

        This method automatically creates the database, collection, and vector
        search index if they don't exist.

        Args:
            documents (`list[Document]`):
                List of Document objects to insert.
            **kwargs (`Any`):
                Additional arguments (unused).

        .. note::
            Each inserted record has structure:

            .. code-block:: python

                {
                    "id": str,                # Document ID
                    "vector": list[float],    # Vector embedding
                    "payload": dict,          # DocMetadata as dict
                }
        """
        await self._validate_db_and_collection()

        # Prepare documents for insertion
        docs_to_insert = []
        for doc in documents:
            # Convert DocMetadata to dict for storage
            payload = {
                "doc_id": doc.metadata.doc_id,
                "chunk_id": doc.metadata.chunk_id,
                "total_chunks": doc.metadata.total_chunks,
                "content": doc.metadata.content,
            }

            # Create document record
            doc_record = {
                "id": f"{doc.metadata.doc_id}_{doc.metadata.chunk_id}",
                "vector": doc.embedding,
                "payload": payload,
            }
            docs_to_insert.append(doc_record)

        # Insert documents using upsert to handle duplicates
        if not docs_to_insert:
            return
        from pymongo import ReplaceOne

        operations = [
            ReplaceOne(
                {"id": doc_record["id"]},
                doc_record,
                upsert=True,
            )
            for doc_record in docs_to_insert
        ]
        await self._collection.bulk_write(operations)

    async def search(
        self,
        query_embedding: Embedding,
        limit: int,
        score_threshold: float | None = None,
        **kwargs: Any,
    ) -> list[Document]:
        """Search relevant documents using MongoDB Vector Search.

        This method uses MongoDB's $vectorSearch aggregation pipeline for
        vector similarity search. It automatically waits for the vector search
        index to be ready before performing the search.

        Args:
            query_embedding (`Embedding`):
                The embedding vector to search for.
            limit (`int`):
                Maximum number of documents to return.
            score_threshold (`float | None`, optional):
                Minimum similarity score threshold. Documents with scores below
                this threshold will be filtered out.
            **kwargs (`Any`):
                Additional arguments for the search operation.

        Returns:
            `list[Document]`: List of Document objects with embedding,
            score, and metadata.

        .. note::
            - Requires MongoDB with vector search support
            - Uses $vectorSearch aggregation pipeline
        """
        await self._validate_db_and_collection()
        # Wait for index to be ready before searching
        await self._wait_for_index_ready()

        # Construct aggregation pipeline for vector search
        # See: https://www.mongodb.com/docs/atlas/atlas-search/vector-search/
        num_candidates = int(
            kwargs.pop(
                "num_candidates",
                max(
                    100,
                    limit * 20,
                ),
            ),
        )

        pipeline: list[dict[str, Any]] = [
            {
                "$vectorSearch": {
                    "index": self.index_name,
                    "path": "vector",
                    "queryVector": list(query_embedding),
                    "numCandidates": num_candidates,
                    "limit": limit,
                    **kwargs,
                },
            },
            {
                "$project": {
                    "vector": 1,
                    "payload": 1,
                    "score": {"$meta": "vectorSearchScore"},
                },
            },
        ]

        cursor = await self._collection.aggregate(pipeline)
        results: list[Document] = []
        async for item in cursor:
            score_val = float(item.get("score", 0.0))
            if score_threshold is not None and score_val < score_threshold:
                continue

            payload = item.get("payload", {})
            # Rebuild Document
            metadata = DocMetadata(**payload)

            results.append(
                Document(
                    embedding=[float(x) for x in item.get("vector", [])],
                    score=score_val,
                    metadata=metadata,
                ),
            )

        return results

    async def delete(
        self,
        ids: str | list[str] | None = None,
    ) -> None:
        """Delete documents from the MongoDB collection.

        Args:
            ids (`str | list[str] | None`, optional):
                List of document IDs to delete. If provided, deletes documents
                with matching doc_id in payload.
        """

        if not ids:
            return

        if isinstance(ids, str):
            ids = [ids]

        await self._collection.delete_many({"payload.doc_id": {"$in": ids}})

    def get_client(self) -> AsyncMongoClient:
        """Get the underlying MongoDB client for advanced operations.

        Returns:
            `AsyncMongoClient`: The AsyncMongoClient instance.
        """
        return self._client

    async def delete_collection(self) -> None:
        """Delete the entire collection.

        .. warning::
            This will permanently delete all documents in the collection.
        """
        await self._collection.drop()

    async def delete_database(self) -> None:
        """Delete the entire database.

        .. warning::
            This will permanently delete the entire database and all its
            collections.
        """
        await self._client.drop_database(self.db_name)

    async def close(self) -> None:
        """Close the MongoDB connection.

        This should be called when the store is no longer needed to properly
        clean up resources.
        """
        await self._client.close()



# -*- coding: utf-8 -*-
"""The A2A agent card resolver base class."""
from abc import abstractmethod
from typing import Any, TYPE_CHECKING

if TYPE_CHECKING:
    from a2a.types import AgentCard
else:
    AgentCard = "a2a.types.AgentCard"


class AgentCardResolverBase:
    """Base class for A2A agent card resolvers, responsible for fetching
    agent cards from various sources. Implementations must provide the
    `get_agent_card` method to retrieve the agent card.
    """

    @abstractmethod
    async def get_agent_card(self, *args: Any, **kwargs: Any) -> AgentCard:
        """Get Agent Card from the configured source.

        Returns:
            `AgentCard`:
                The resolved agent card object.
        """



# -*- coding: utf-8 -*-
"""Utils for agents in agentscope."""
from typing import Any


class _AsyncNullContext:
    """An async null context manager."""

    async def __aenter__(self) -> None:
        return None

    async def __aexit__(
        self,
        exc_type: Any,
        exc_val: Any,
        exc_tb: Any,
    ) -> None:
        return None



# -*- coding: utf-8 -*-
"""The Redis session class."""
import json
from typing import Any, TYPE_CHECKING

from ._session_base import SessionBase
from .._logging import logger
from ..module import StateModule

if TYPE_CHECKING:
    from redis.asyncio import ConnectionPool, Redis
else:
    ConnectionPool = Any
    Redis = Any


class RedisSession(SessionBase):
    """The Redis session class."""

    SESSION_KEY = "user_id:{user_id}:session:{session_id}:state"
    """Redis key pattern (without prefix) for storing sessions state for
    a specific session.
    """

    def __init__(
        self,
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: str | None = None,
        connection_pool: ConnectionPool | None = None,
        key_ttl: int | None = None,
        key_prefix: str = "",
        **kwargs: Any,
    ) -> None:
        """Initialize the Redis session class by connecting to Redis.

        Args:
            host (`str`, defaults to `"localhost"`):
                Redis server host.
            port (`int`, defaults to `6379`):
                Redis server port.
            db (`int`, defaults to `0`):
                Redis database index.
            password (`str | None`, optional):
                Redis password if required.
            connection_pool (`ConnectionPool | None`, optional):
                Optional Redis connection pool.
            key_ttl (`int | None`, optional):
                Expire time in seconds for each session state key. If provided,
                the expiration will be refreshed on every save/load
                (sliding TTL). If `None`, the session state will not expire.
            key_prefix (`str`, default to `""`):
                Optional Redis key prefix prepended to every key generated by
                this storage. Useful for isolating keys across
                apps/environments (e.g. `"prod:"`, `"staging:"`, `"myapp:"`).
            **kwargs (`Any`):
                Additional keyword arguments passed to redis client.
        """
        try:
            import redis.asyncio as redis
        except ImportError as e:
            raise ImportError(
                "The 'redis' package is required for RedisSession. "
                "Please install it via 'pip install redis[async]'.",
            ) from e

        self.key_ttl = key_ttl
        self.key_prefix = key_prefix or ""

        self._client = redis.Redis(
            host=host,
            port=port,
            db=db,
            password=password,
            connection_pool=connection_pool,
            decode_responses=True,
            **kwargs,
        )

    def get_client(self) -> Redis:
        """Get the underlying Redis client.

        Returns:
            `Redis`:
                The Redis client instance.
        """
        return self._client

    def _get_session_key(
        self,
        session_id: str,
        user_id: str,
    ) -> str:
        """Get the Redis key to store a given session state."""
        return self.key_prefix + self.SESSION_KEY.format(
            user_id=user_id,
            session_id=session_id,
        )

    async def save_session_state(
        self,
        session_id: str,
        user_id: str = "default_user",
        **state_modules_mapping: StateModule,
    ) -> None:
        """Save the state dictionary to Redis.

        Args:
            session_id (`str`):
                The session id.
            user_id (`str`, default to `"default_user"`):
                The user ID for the storage.
            **state_modules_mapping (`dict[str, StateModule]`):
                A dictionary mapping of state module names to their instances.
        """
        state_dicts = {
            name: state_module.state_dict()
            for name, state_module in state_modules_mapping.items()
        }

        key = self._get_session_key(session_id, user_id=user_id)
        value = json.dumps(state_dicts, ensure_ascii=False)

        await self._client.set(key, value, ex=self.key_ttl)

        logger.info("Save session state to redis key %s successfully.", key)

    async def load_session_state(
        self,
        session_id: str,
        user_id: str = "default_user",
        allow_not_exist: bool = True,
        **state_modules_mapping: StateModule,
    ) -> None:
        """Load the state dictionary from Redis.

        Args:
            session_id (`str`):
                The session id.
            user_id (`str`, default to `"default_user"`):
                The user ID for the storage.
            allow_not_exist (`bool`, defaults to `True`):
                Whether to allow the session to not exist.
            **state_modules_mapping (`dict[str, StateModule]`):
                The mapping of state modules to be loaded.
        """
        key = self._get_session_key(session_id, user_id=user_id)

        # Use GETEX to get and refresh TTL in a single atomic operation
        if self.key_ttl is not None:
            data = await self._client.getex(key, ex=self.key_ttl)
        else:
            data = await self._client.get(key)

        if data is None:
            if allow_not_exist:
                logger.info(
                    "Session key %s does not exist in Redis. Skip loading "
                    "session state.",
                    key,
                )
                return
            raise ValueError(
                f"Failed to load session state because redis key {key} "
                "does not exist.",
            )

        if isinstance(data, (bytes, bytearray)):
            data = data.decode("utf-8")

        states = json.loads(data)

        for name, state_module in state_modules_mapping.items():
            if name in states:
                state_module.load_state_dict(states[name])

        logger.info("Load session state from redis key %s successfully.", key)

    async def close(self) -> None:
        """Close the Redis client connection."""
        await self._client.close()

    async def __aenter__(self) -> "RedisSession":
        """Enter the async context manager.

        Returns:
            `RedisSession`:
                The current `RedisSession` instance.
        """
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        """Exit the async context manager and close the connection.

        Args:
            exc_type (`type[BaseException] | None`):
                The type of the exception.
            exc_value (`BaseException | None`):
                The exception instance.
            traceback (`Any`):
                The traceback.
        """
        await self.close()



# -*- coding: utf-8 -*-
"""The redis based memory storage implementation."""
import json
from typing import Any, TYPE_CHECKING

from ._base import MemoryBase
from ...message import Msg

if TYPE_CHECKING:
    from redis.asyncio import ConnectionPool, Redis
else:
    ConnectionPool = Any
    Redis = Any


class RedisMemory(MemoryBase):
    """Redis memory storage implementation, which supports session and user
    context.

    .. note:: All the operations in this class are within a specific session
     and user context, identified by `session_id` and `user_id`. Cross-session
     or cross-user operations are not supported. For example, the
     `remove_messages` method will only remove messages that belong to the
     specified `session_id` and `user_id`.

    .. note:: All Redis keys used by this class will be prefixed by `prefix`
    (if provided) to support multi-tenant / multi-app isolation.

    **Mark Index Storage:**

    This class maintains a `marks_index` (Redis Set) to efficiently track all
    mark names within a session. When a mark is created via `add_mark()`, the
    mark name is added to this set. This allows quick retrieval of all marks
    without scanning all Redis keys. The marks_index key pattern is:
    ``user_id:{user_id}:session:{session_id}:marks_index``

    Each individual mark stores its associated message IDs in a separate Redis
    List with the key pattern:
    ``user_id:{user_id}:session:{session_id}:mark:{mark}``

    """

    SESSION_KEY = "user_id:{user_id}:session:{session_id}:messages"
    """Redis key pattern (without prefix) for storing message IDs (ordered) for
    a specific session.
    """

    SESSION_PATTERN = "user_id:{user_id}:session:{session_id}:*"
    """Redis key pattern (without prefix) for scanning all keys belong to
    a specific user and session."""

    MARK_KEY = "user_id:{user_id}:session:{session_id}:mark:{mark}"
    """Redis key pattern (without prefix) for storing message IDs that belong
    to a specific mark.
    """

    MESSAGE_KEY = "user_id:{user_id}:session:{session_id}:msg:{msg_id}"
    """Redis key pattern (without prefix) for storing message payload data."""

    MARKS_INDEX_KEY = "user_id:{user_id}:session:{session_id}:marks_index"
    """Redis key pattern (without prefix) for storing all mark names as a set.
    This is used to avoid scanning all keys to find marks.
    """

    def __init__(
        self,
        session_id: str = "default_session",
        user_id: str = "default_user",
        host: str = "localhost",
        port: int = 6379,
        db: int = 0,
        password: str | None = None,
        connection_pool: ConnectionPool | None = None,
        key_prefix: str = "",
        key_ttl: int | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Redis based storage by connecting to the Redis
        server. You can provide either the connection parameters or an
        existing connection pool.

        Args:
            session_id (`str`, default to `"default_session"`):
                The session ID for the storage.
            user_id (`str`, default to `"default_user"`):
                The user ID for the storage.
            host (`str`, default to `"localhost"`):
                The Redis server host.
            port (`int`, default to `6379`):
                The Redis server port.
            db (`int`, default to `0`):
                The Redis database index.
            password (`str | None`, optional):
                The password for the Redis server, if required.
            connection_pool (`ConnectionPool | None`, optional):
                An optional Redis connection pool. If provided, it will be used
                instead of creating a new connection.
            key_prefix (`str`, default to `""`):
                Optional Redis key prefix prepended to every key generated by
                this storage. Useful for isolating keys across
                apps/environments (e.g. `"prod"`, `"staging"`, `"myapp"`).
            key_ttl (`int | None`, default to `None`):
                The expired time in seconds for each key. If provided, the
                expiration will be refreshed on every access (sliding TTL). If
                `None`, the keys will not expire. **Note** the ttl will update
                all session related keys, so it's not recommended to set for
                large sessions.
            **kwargs (`Any`):
                Additional keyword arguments to pass to the Redis client.
        """
        try:
            import redis.asyncio as redis
        except ImportError as e:
            raise ImportError(
                "The 'redis' package is required for RedisStorage. "
                "Please install it via 'pip install redis[async]'.",
            ) from e

        super().__init__()

        self.session_id = session_id
        self.user_id = user_id
        self.key_prefix = key_prefix or ""
        self.key_ttl = key_ttl

        self._client = redis.Redis(
            host=host,
            port=port,
            db=db,
            password=password,
            connection_pool=connection_pool,
            decode_responses=True,
            **kwargs,
        )

    def get_client(self) -> Redis:
        """Get the underlying Redis client.

        Returns:
            `Redis`:
                The Redis client instance.
        """
        return self._client

    def _decode_if_bytes(self, data: Any) -> Any:
        """Helper method to decode bytes to str if needed.

        Args:
            data (`Any`):
                The data to decode, which may be bytes, bytearray, or str.

        Returns:
            `Any`:
                The decoded string if input was bytes/bytearray, otherwise
                the original data.
        """
        if isinstance(data, (bytes, bytearray)):
            return data.decode("utf-8")
        return data

    def _decode_list(self, data_list: list) -> list:
        """Helper method to decode a list of potential bytes.

        Args:
            data_list (`list`):
                A list that may contain bytes, bytearray, or str elements.

        Returns:
            `list`:
                A list with all bytes/bytearray elements decoded to str.
        """
        return [self._decode_if_bytes(item) for item in data_list]

    def _get_session_key(self) -> str:
        """Get the Redis key for the current session.

        Returns:
            `str`:
                The Redis key for storing messages in the current session.
        """
        return self.key_prefix + self.SESSION_KEY.format(
            user_id=self.user_id,
            session_id=self.session_id,
        )

    def _get_session_pattern(self) -> str:
        """Get the Redis key pattern for all keys in the current session.

        Returns:
            `str`:
                The Redis key pattern for all keys in the current session.
        """
        return self.key_prefix + self.SESSION_PATTERN.format(
            user_id=self.user_id,
            session_id=self.session_id,
        )

    def _get_mark_key(self, mark: str) -> str:
        """Get the Redis key for a specific mark.

        Args:
            mark (`str`):
                The mark name.

        Returns:
            `str`:
                The Redis key for storing message IDs with the given mark.
        """
        return self.key_prefix + self.MARK_KEY.format(
            user_id=self.user_id,
            session_id=self.session_id,
            mark=mark,
        )

    def _get_mark_pattern(self) -> str:
        """Get the Redis key pattern for all marks in the current session.

        Returns:
            `str`:
                The Redis key pattern for all mark keys.
        """
        return self.key_prefix + self.MARK_KEY.format(
            user_id=self.user_id,
            session_id=self.session_id,
            mark="*",
        )

    def _get_marks_index_key(self) -> str:
        """Get the Redis key for the marks index set.

        Returns:
            `str`:
                The Redis key for storing all mark names as a set.
        """
        return self.key_prefix + self.MARKS_INDEX_KEY.format(
            user_id=self.user_id,
            session_id=self.session_id,
        )

    def _extract_mark_from_key(self, mark_key: str) -> str:
        """Extract the mark name from a full mark key.

        Args:
            mark_key (`str`):
                The full Redis key for a mark.

        Returns:
            `str`:
                The mark name extracted from the key.
        """
        # Remove the prefix and the base pattern to get the mark name
        # Example: "prefix:user_id:xxx:session:yyy:mark:my_mark" -> "my_mark"
        prefix_pattern = self.key_prefix + self.MARK_KEY.format(
            user_id=self.user_id,
            session_id=self.session_id,
            mark="",
        )
        return mark_key.replace(prefix_pattern, "")

    def _get_message_key(self, msg_id: str) -> str:
        """Get the Redis key for a specific message.

        Args:
            msg_id (`str`):
                The message ID.

        Returns:
            `str`:
                The Redis key for storing the message data.
        """
        return self.key_prefix + self.MESSAGE_KEY.format(
            user_id=self.user_id,
            session_id=self.session_id,
            msg_id=msg_id,
        )

    async def _refresh_session_ttl(
        self,
        pipe: Any | None = None,
    ) -> None:
        """Refresh the TTL for the session keys (if `key_ttl` is set).

        Args:
            pipe (`Any | None`, optional):
                An optional Redis pipeline to use. If `None`, a new pipeline
                will be created and executed immediately. If provided, the
                expired commands will be added to the pipeline without
                executing it.
        """
        if self.key_ttl is None:
            return

        # Create a new pipeline if not provided
        should_execute = pipe is None
        if pipe is None:
            pipe = self._client.pipeline()

        cursor = 0
        while True:
            cursor, keys = await self._client.scan(
                cursor,
                match=self._get_session_pattern(),
                count=100,
            )
            # Decode keys if they are bytes
            keys = self._decode_list(keys)
            for key in keys:
                await pipe.expire(key, self.key_ttl)
            if cursor == 0:
                break

        if should_execute:
            await pipe.execute()

    async def _scan_and_migrate_marks(self) -> list[str]:
        """Scan all mark keys and migrate them to the marks index.

        This method is only called once for old data that doesn't have
        a marks index yet. After migration, the marks index will be
        maintained automatically.

        Returns:
            `list[str]`:
                The list of all mark keys found.
        """
        mark_keys = []
        cursor = 0
        while True:
            cursor, keys = await self._client.scan(
                cursor,
                match=self._get_mark_pattern(),
                count=50,
            )
            keys = self._decode_list(keys)
            mark_keys.extend(keys)
            if cursor == 0:
                break

        # Build the marks index
        if mark_keys:
            pipe = self._client.pipeline()
            for mark_key in mark_keys:
                mark = self._extract_mark_from_key(mark_key)
                await pipe.sadd(self._get_marks_index_key(), mark)
            await pipe.execute()

        return mark_keys

    async def _get_all_mark_keys(self) -> list[str]:
        """Get all mark keys, compatible with both old and new data.

        For new data (with marks index), this method uses the index directly.
        For old data (without marks index), this method scans once and
        migrates to the new structure.

        Returns:
            `list[str]`:
                The list of all mark keys.
        """
        marks_index_key = self._get_marks_index_key()

        # Try to read from the index first
        marks = await self._client.smembers(marks_index_key)
        if marks:
            # Index exists, use it
            marks = self._decode_list(list(marks))
            return [self._get_mark_key(mark) for mark in marks]

        # Index doesn't exist, check if this is a new session
        session_exists = await self._client.exists(self._get_session_key())
        if not session_exists:
            # New session, no data at all, return empty
            return []

        # Old session without index, need to scan and migrate (only once)
        mark_keys = await self._scan_and_migrate_marks()
        return mark_keys

    async def get_memory(
        self,
        mark: str | None = None,
        exclude_mark: str | None = None,
        prepend_summary: bool = True,
        **kwargs: Any,
    ) -> list[Msg]:
        """Get the messages from the memory by mark (if provided). Otherwise,
        get all messages.

        .. note:: If `mark` and `exclude_mark` are both provided, the messages
         will be filtered by both arguments.

        .. note:: `mark` and `exclude_mark` should not overlap.

        Args:
            mark (`str | None`, optional):
                The mark to filter messages. If `None`, return all messages.
            exclude_mark (`str | None`, optional):
                The mark to exclude messages. If provided, messages with
                this mark will be excluded from the results.
            prepend_summary (`bool`, defaults to True):
                Whether to prepend the compressed summary as a message

        Returns:
            `list[Msg]`:
                The list of messages retrieved from the storage.
        """
        # Type checks
        if not (mark is None or isinstance(mark, str)):
            raise TypeError(
                f"The mark should be a string or None, but got {type(mark)}.",
            )

        if not (exclude_mark is None or isinstance(exclude_mark, str)):
            raise TypeError(
                f"The exclude_mark should be a string or None, but got "
                f"{type(exclude_mark)}.",
            )

        if mark is None:
            # Obtain the message IDs from the session list
            msg_ids = await self._client.lrange(self._get_session_key(), 0, -1)

        else:
            # Obtain the message IDs from the mark list
            msg_ids = await self._client.lrange(
                self._get_mark_key(mark),
                0,
                -1,
            )

        msg_ids = self._decode_list(msg_ids)

        # Exclude messages by exclude_mark
        if exclude_mark:
            exclude_msg_ids = await self._client.lrange(
                self._get_mark_key(exclude_mark),
                0,
                -1,
            )
            exclude_msg_ids = self._decode_list(exclude_msg_ids)
            msg_ids = [_ for _ in msg_ids if _ not in exclude_msg_ids]

        # Use mget for batch retrieval to avoid N+1 queries
        messages: list[Msg] = []
        if msg_ids:
            msg_keys = [self._get_message_key(msg_id) for msg_id in msg_ids]
            msg_data_list = await self._client.mget(msg_keys)

            for msg_data in msg_data_list:
                if msg_data is not None:
                    # Decode if bytes
                    msg_data = self._decode_if_bytes(msg_data)
                    msg_dict = json.loads(msg_data)
                    messages.append(Msg.from_dict(msg_dict))

        # Refresh TTLs
        await self._refresh_session_ttl()

        if prepend_summary and self._compressed_summary:
            return [
                Msg(
                    "user",
                    self._compressed_summary,
                    "user",
                ),
                *messages,
            ]

        return messages

    async def add(
        self,
        memories: Msg | list[Msg] | None,
        marks: str | list[str] | None = None,
        skip_duplicated: bool = True,
        **kwargs: Any,
    ) -> None:
        """Add message into the storage with the given mark (if provided).

        Args:
            memories (`Msg | list[Msg]`):
                The message(s) to be added.
            marks (`str | list[str] | None`, optional):
                The mark(s) to associate with the message(s). If `None`, no
                mark is associated.
            skip_duplicated (`bool`, defaults to `True`):
                If `True`, skip messages with duplicate IDs that already exist
                in the storage. If `False`, allow duplicate message IDs to be
                added to the session list (though the message data will be
                overwritten).
        """
        if memories is None:
            return

        if isinstance(memories, Msg):
            memories = [memories]

        # Normalize marks to a list
        if marks is None:
            mark_list = []
        elif isinstance(marks, str):
            mark_list = [marks]
        else:
            mark_list = marks

        # Filter out existing messages if skip_duplicated is True
        messages_to_add = memories
        if skip_duplicated:
            # Get all existing message IDs in the current session
            existing_msg_ids = await self._client.lrange(
                self._get_session_key(),
                0,
                -1,
            )
            existing_msg_ids = self._decode_list(existing_msg_ids)
            existing_msg_ids_set = set(existing_msg_ids)

            # Filter out messages that already exist
            messages_to_add = [
                m for m in memories if m.id not in existing_msg_ids_set
            ]

            # If all messages are duplicates, return early
            if not messages_to_add:
                return

        # Use pipeline for atomic operations
        pipe = self._client.pipeline()

        # Push message ids into the session list
        if messages_to_add:
            await pipe.rpush(
                self._get_session_key(),
                *[m.id for m in messages_to_add],
            )

        # Store message data and marks
        for m in messages_to_add:
            # Record the message data
            await pipe.set(
                self._get_message_key(m.id),
                json.dumps(m.to_dict(), ensure_ascii=False),
            )

            # Record the marks if provided
            for mark in mark_list:
                await pipe.rpush(self._get_mark_key(mark), m.id)
                # Maintain the marks index
                await pipe.sadd(self._get_marks_index_key(), mark)

        # Refresh TTLs
        await self._refresh_session_ttl(pipe=pipe)

        await pipe.execute()

    async def delete(
        self,
        msg_ids: list[str],
        **kwargs: Any,
    ) -> int:
        """Remove message(s) from the storage by their IDs.

        Args:
            msg_ids (`list[str]`):
                The list of message IDs to be removed.

        Returns:
            `int`:
                The number of messages removed.
        """
        if not msg_ids:
            return 0

        # Get all mark keys using the new method (compatible with old data)
        mark_keys = await self._get_all_mark_keys()

        pipe = self._client.pipeline()
        for msg_id in msg_ids:
            # Remove from the session (0 means remove all occurrences)
            await pipe.lrem(self._get_session_key(), 0, msg_id)

            # Remove the message data
            await pipe.delete(self._get_message_key(msg_id))

            # Remove from all marks
            for mark_key in mark_keys:
                await pipe.lrem(mark_key, 0, msg_id)

        # Refresh TTLs
        await self._refresh_session_ttl(pipe=pipe)

        results = await pipe.execute()

        # Count actual deletions from lrem results (every 3rd result
        # starting from 0)
        removed_count = sum(
            1
            for i in range(
                0,
                len(msg_ids) * (2 + len(mark_keys)),
                2 + len(mark_keys),
            )
            if results[i] > 0
        )

        return removed_count

    async def delete_by_mark(
        self,
        mark: str | list[str],
        **kwargs: Any,
    ) -> int:
        """Remove messages from the storage by their marks.

        Args:
            mark (`str | list[str]`):
                The mark(s) of the messages to be removed.

        Returns:
            `int`:
                The number of messages removed.
        """
        if isinstance(mark, str):
            mark = [mark]

        total_removed = 0

        for m in mark:
            mark_key = self._get_mark_key(m)
            msg_ids = await self._client.lrange(mark_key, 0, -1)
            msg_ids = self._decode_list(msg_ids)

            if not msg_ids:
                continue

            # Remove messages by IDs
            removed_count = await self.delete(
                msg_ids,
            )
            total_removed += removed_count

            # Delete the mark list
            await self._client.delete(mark_key)

            # Remove from the marks index
            await self._client.srem(self._get_marks_index_key(), m)

        # Refresh TTLs
        await self._refresh_session_ttl()

        return total_removed

    async def clear(self) -> None:
        """Clear all messages belong to this session from the storage."""
        msg_ids = await self._client.lrange(self._get_session_key(), 0, -1)
        msg_ids = self._decode_list(msg_ids)

        # Get all mark keys using the new method (compatible with old data)
        mark_keys = await self._get_all_mark_keys()

        pipe = self._client.pipeline()

        for msg_id in msg_ids:
            # Remove the message data
            await pipe.delete(self._get_message_key(msg_id))

        # Delete the session list
        await pipe.delete(self._get_session_key())

        # Delete all mark lists
        for mark_key in mark_keys:
            await pipe.delete(mark_key)

        # Delete the marks index
        await pipe.delete(self._get_marks_index_key())

        await pipe.execute()

    async def size(self) -> int:
        """Get the number of messages in the storage.

        Returns:
            `int`:
                The number of messages in the storage.
        """
        size = await self._client.llen(self._get_session_key())
        await self._refresh_session_ttl()
        return size

    async def update_messages_mark(
        self,
        new_mark: str | None,
        old_mark: str | None = None,
        msg_ids: list[str] | None = None,
    ) -> int:
        """A unified method to update marks of messages in the storage (add,
        remove, or change marks).

        - If `msg_ids` is provided, the update will be applied to the messages
         with the specified IDs.
        - If `old_mark` is provided, the update will be applied to the
         messages with the specified old mark. Otherwise, the `new_mark` will
         be added to all messages (or those filtered by `msg_ids`).
        - If `new_mark` is `None`, the mark will be removed from the messages.

        Args:
            new_mark (`str | None`, optional):
                The new mark to set for the messages. If `None`, the mark
                will be removed.
            old_mark (`str | None`, optional):
                The old mark to filter messages. If `None`, this constraint
                is ignored.
            msg_ids (`list[str] | None`, optional):
                The list of message IDs to be updated. If `None`, this
                constraint is ignored.

        Returns:
            `int`:
                The number of messages updated.
        """
        # Determine which message IDs to update
        # Get source key based on old_mark
        source_key = (
            self._get_mark_key(old_mark)
            if old_mark is not None
            else self._get_session_key()
        )
        mark_msg_ids = await self._client.lrange(source_key, 0, -1)
        mark_msg_ids = self._decode_list(mark_msg_ids)

        # Check if we're removing all messages from old_mark
        removing_all_from_old_mark = old_mark is not None and (
            msg_ids is None or all(mid in set(msg_ids) for mid in mark_msg_ids)
        )

        # Filter by msg_ids if provided
        if msg_ids is not None:
            msg_ids_set = set(msg_ids)
            mark_msg_ids = [mid for mid in mark_msg_ids if mid in msg_ids_set]

        if not mark_msg_ids:
            return 0

        # Get existing IDs in new_mark list once (if needed)
        existing_ids_set = set()
        new_mark_key = None
        if new_mark is not None:
            new_mark_key = self._get_mark_key(new_mark)
            existing_ids = await self._client.lrange(new_mark_key, 0, -1)
            existing_ids = self._decode_list(existing_ids)
            existing_ids_set = set(existing_ids)

        # Use pipeline for batch operations
        pipe = self._client.pipeline()
        updated_count = 0

        for msg_id in mark_msg_ids:
            # Remove from old_mark list if applicable
            if old_mark is not None:
                await pipe.lrem(
                    self._get_mark_key(old_mark),
                    0,
                    msg_id,
                )

            # Add to new_mark list only if not already present
            if new_mark is not None and msg_id not in existing_ids_set:
                await pipe.rpush(new_mark_key, msg_id)
                existing_ids_set.add(msg_id)
                # Maintain the marks index
                await pipe.sadd(self._get_marks_index_key(), new_mark)

            # Count update only if we actually did something
            if old_mark is not None or new_mark is not None:
                updated_count += 1

        # Clean up old_mark only if we removed ALL messages from it
        if old_mark is not None and removing_all_from_old_mark:
            old_mark_key = self._get_mark_key(old_mark)
            # After lrem operations, the old mark list will be empty
            # Delete the mark key and remove from index
            await pipe.delete(old_mark_key)
            await pipe.srem(self._get_marks_index_key(), old_mark)

        await self._refresh_session_ttl(pipe=pipe)

        await pipe.execute()
        return updated_count

    async def close(self, close_connection_pool: bool | None = None) -> None:
        """Close the Redis client connection.

        Args:
            close_connection_pool (`bool | None`, optional):
                Decides whether to close the connection pool used by this
                Redis client, overriding Redis.auto_close_connection_pool.
                By default, let Redis.auto_close_connection_pool decide
                whether to close the connection pool
        """
        await self._client.aclose(close_connection_pool=close_connection_pool)

    async def __aenter__(self) -> "RedisMemory":
        """Enter the async context manager.

        Returns:
            `RedisMemory`:
                The memory instance itself.
        """
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: Any,
    ) -> None:
        """Exit the async context manager and close the session.

        Args:
            exc_type (`type[BaseException] | None`):
                The exception type if an exception was raised.
            exc_value (`BaseException | None`):
                The exception instance if an exception was raised.
            traceback (`Any`):
                The traceback object if an exception was raised.
        """
        await self.close()



# -*- coding: utf-8 -*-
"""The in-memory storage module for memory storage."""
from copy import deepcopy
from typing import Any

from ...message import Msg
from ._base import MemoryBase


class InMemoryMemory(MemoryBase):
    """The in-memory implementation of memory storage."""

    def __init__(self) -> None:
        """Initialize the in-memory storage."""
        super().__init__()
        # Use a list of tuples to store messages along with their marks
        self.content: list[tuple[Msg, list[str]]] = []

        # Register the state for serialization
        self.register_state("content")

    async def get_memory(
        self,
        mark: str | None = None,
        exclude_mark: str | None = None,
        prepend_summary: bool = True,
        **kwargs: Any,
    ) -> list[Msg]:
        """Get the messages from the memory by mark (if provided). Otherwise,
        get all messages.

        .. note:: If `mark` and `exclude_mark` are both provided, the messages
         will be filtered by both arguments.

        .. note:: `mark` and `exclude_mark` should not overlap.

        Args:
            mark (`str | None`, optional):
                The mark to filter messages. If `None`, return all messages.
            exclude_mark (`str | None`, optional):
                The mark to exclude messages. If provided, messages with
                this mark will be excluded from the results.
            prepend_summary (`bool`, defaults to True):
                Whether to prepend the compressed summary as a message

        Raises:
            `TypeError`:
                If the provided mark is not a string or None.

        Returns:
            `list[Msg]`:
                The list of messages retrieved from the storage.
        """
        # Type checks
        if not (mark is None or isinstance(mark, str)):
            raise TypeError(
                f"The mark should be a string or None, but got {type(mark)}.",
            )

        if not (exclude_mark is None or isinstance(exclude_mark, str)):
            raise TypeError(
                f"The exclude_mark should be a string or None, but got "
                f"{type(exclude_mark)}.",
            )

        # Filter messages based on mark
        filtered_content = [
            (msg, marks)
            for msg, marks in self.content
            if mark is None or mark in marks
        ]

        # Further filter messages based on exclude_mark
        if exclude_mark is not None:
            filtered_content = [
                (msg, marks)
                for msg, marks in filtered_content
                if exclude_mark not in marks
            ]

        if prepend_summary and self._compressed_summary:
            return [
                Msg(
                    "user",
                    self._compressed_summary,
                    "user",
                ),
                *[msg for msg, _ in filtered_content],
            ]

        return [msg for msg, _ in filtered_content]

    async def add(
        self,
        memories: Msg | list[Msg] | None,
        marks: str | list[str] | None = None,
        allow_duplicates: bool = False,
        **kwargs: Any,
    ) -> None:
        """Add message(s) into the memory storage with the given mark
        (if provided).

        Args:
            memories (`Msg | list[Msg] | None`):
                The message(s) to be added.
            marks (`str | list[str] | None`, optional):
                The mark(s) to associate with the message(s). If `None`, no
                mark is associated.
            allow_duplicates (`bool`, defaults to `False`):
                Whether to allow duplicate messages in the storage.
        """
        if memories is None:
            return

        if isinstance(memories, Msg):
            memories = [memories]

        if marks is None:
            marks = []
        elif isinstance(marks, str):
            marks = [marks]
        elif not isinstance(marks, list) or not all(
            isinstance(m, str) for m in marks
        ):
            raise TypeError(
                f"The mark should be a string, a list of strings, or None, "
                f"but got {type(marks)}.",
            )

        if not allow_duplicates:
            existing_ids = {msg.id for msg, _ in self.content}
            memories = [msg for msg in memories if msg.id not in existing_ids]

        for msg in memories:
            self.content.append((deepcopy(msg), deepcopy(marks)))

    async def delete(
        self,
        msg_ids: list[str],
        **kwargs: Any,
    ) -> int:
        """Remove message(s) from the storage by their IDs.

        Args:
            msg_ids (`list[str]`):
                The list of message IDs to be removed.

        Returns:
            `int`:
                The number of messages removed.
        """
        initial_size = len(self.content)
        self.content = [
            (msg, marks)
            for msg, marks in self.content
            if msg.id not in msg_ids
        ]
        return initial_size - len(self.content)

    async def delete_by_mark(
        self,
        mark: str | list[str],
        **kwargs: Any,
    ) -> int:
        """Remove messages from the memory by their marks.

        Args:
            mark (`str | list[str]`):
                The mark(s) of the messages to be removed.

        Raises:
            `TypeError`:
                If the provided mark is not a string or a list of strings.

        Returns:
            `int`:
                The number of messages removed.
        """
        if isinstance(mark, str):
            mark = [mark]

        if isinstance(mark, list) and not all(
            isinstance(m, str) for m in mark
        ):
            raise TypeError(
                f"The mark should be a string or a list of strings, "
                f"but got {type(mark)} with elements of types "
                f"{[type(m) for m in mark]}.",
            )

        initial_size = len(self.content)
        for m in mark:
            self.content = [
                (msg, marks) for msg, marks in self.content if m not in marks
            ]

        return initial_size - len(self.content)

    async def clear(self) -> None:
        """Clear all messages from the storage."""
        self.content.clear()

    async def size(self) -> int:
        """Get the number of messages in the storage.

        Returns:
            `int`:
                The number of messages in the storage.
        """
        return len(self.content)

    async def update_messages_mark(
        self,
        new_mark: str | None,
        old_mark: str | None = None,
        msg_ids: list[str] | None = None,
    ) -> int:
        """A unified method to update marks of messages in the storage (add,
        remove, or change marks).

        - If `msg_ids` is provided, the update will be applied to the messages
         with the specified IDs.
        - If `old_mark` is provided, the update will be applied to the
         messages with the specified old mark. Otherwise, the `new_mark` will
         be added to all messages (or those filtered by `msg_ids`).
        - If `new_mark` is `None`, the mark will be removed from the messages.

        Args:
            new_mark (`str | None`, optional):
                The new mark to set for the messages. If `None`, the mark
                will be removed.
            old_mark (`str | None`, optional):
                The old mark to filter messages. If `None`, this constraint
                is ignored.
            msg_ids (`list[str] | None`, optional):
                The list of message IDs to be updated. If `None`, this
                constraint is ignored.

        Returns:
            `int`:
                The number of messages updated.
        """
        updated_count = 0

        for idx, (msg, marks) in enumerate(self.content):
            # If msg_ids is provided, skip messages not in the list
            if msg_ids is not None and msg.id not in msg_ids:
                continue

            # If old_mark is provided, skip messages that do not have the old
            # mark
            if old_mark is not None and old_mark not in marks:
                continue

            # If new_mark is None, remove the old_mark
            if new_mark is None:
                if old_mark in marks:
                    marks.remove(old_mark)
                    updated_count += 1

            else:
                # If new_mark is provided, add or replace the old_mark
                if old_mark is not None and old_mark in marks:
                    marks.remove(old_mark)
                if new_mark not in marks:
                    marks.append(new_mark)
                    updated_count += 1

            self.content[idx] = (msg, marks)

        return updated_count

    def state_dict(self) -> dict:
        """Get the state dictionary for serialization."""
        return {
            **super().state_dict(),
            "content": [[msg.to_dict(), marks] for msg, marks in self.content],
        }

    def load_state_dict(self, state_dict: dict, strict: bool = True) -> None:
        """Load the state dictionary for deserialization."""
        if strict and "content" not in state_dict:
            raise KeyError(
                "The state_dict does not contain 'content' "
                "keys required for InMemoryMemory.",
            )

        self._compressed_summary = state_dict.get("_compressed_summary", "")

        self.content = []
        for item in state_dict.get("content", []):
            if isinstance(item, (tuple, list)) and len(item) == 2:
                msg_dict, marks = item
                msg = Msg.from_dict(msg_dict)
                self.content.append((msg, marks))

            elif isinstance(item, dict):
                # For compatibility with older versions
                msg = Msg.from_dict(item)
                self.content.append((msg, []))

            else:
                raise ValueError(
                    "Invalid item format in state_dict for InMemoryMemory.",
                )



# -*- coding: utf-8 -*-
"""This module has been deprecated and renamed to 'agentscope.tuner'."""

raise ImportError(
    "The 'agentscope.tune' module has been renamed to 'agentscope.tuner'. "
    "Please update your imports: 'from agentscope.tuner import ...'",
)



# -*- coding: utf-8 -*-
"""The realtime model base class."""
import asyncio
import json
from abc import abstractmethod
from asyncio import Queue
from typing import Any

from ._events import ModelEvents
from ..message import AudioBlock, TextBlock, ImageBlock, ToolResultBlock


class RealtimeModelBase:
    """The realtime model base class."""

    model_name: str
    """The model name"""

    support_input_modalities: list[str]
    """The supported input modalities of the DashScope realtime model."""

    websocket_url: str
    """The websocket URL of the realtime model API."""

    websocket_headers: dict[str, str]
    """The websocket headers of the realtime model API."""

    input_sample_rate: int
    """The input audio sample rate."""

    output_sample_rate: int
    """The output audio sample rate."""

    def __init__(
        self,
        model_name: str,
    ) -> None:
        """Initialize the RealtimeModelBase class.

        Args:
            model_name (`str`):
                The model name.
        """

        self.model_name = model_name

        # The incoming queue to handle the data returned from the realtime
        # model API.
        self._incoming_queue = Queue()
        self._incoming_task = None

        from websockets import ClientConnection

        self._websocket: ClientConnection | None = None

    @abstractmethod
    async def send(
        self,
        data: AudioBlock | TextBlock | ImageBlock | ToolResultBlock,
    ) -> None:
        """Send data to the realtime model for processing.

        Args:
            data (`AudioBlock | TextBlock | ImageBlock | ToolResultBlock`):
                The data to be sent to the realtime model.
        """

    async def connect(
        self,
        outgoing_queue: Queue,
        instructions: str,
        tools: list[dict] | None = None,
    ) -> None:
        """Establish a connection to the realtime model.

        Args:
            outgoing_queue (`Queue`):
                The queue to push the model responses to the outside.
            instructions (`str`):
                The instructions to guide the realtime model's behavior.
            tools (`list[dict]`, *optional*):
                The list of tools JSON schemas.
        """
        import websockets

        self._websocket = await websockets.connect(
            self.websocket_url,
            additional_headers=self.websocket_headers,
        )

        self._incoming_task = asyncio.create_task(
            self._receive_model_event_loop(outgoing_queue),
        )

        # Updating the session with instructions and other configurations
        session_config = self._build_session_config(instructions, tools)
        await self._websocket.send(
            json.dumps(session_config, ensure_ascii=False),
        )

    @abstractmethod
    def _build_session_config(
        self,
        instructions: str,
        tools: list[dict] | None,
        **kwargs: Any,
    ) -> dict:
        """Build the session configuration message to initialize or update
        the realtime model session.

        Args:
            instructions (`str`):
                The instructions to guide the realtime model's behavior.
            tools (`list[dict]`, optional):
                The list of tools available to the realtime model.
            **kwargs (`Any`):
                Additional keyword arguments for session configuration.

        Returns:
            `dict`:
                The session configuration message.
        """

    async def disconnect(self) -> None:
        """Close the connection to the realtime model."""
        # TODO: session ended

        if self._incoming_task and not self._incoming_task.done():
            self._incoming_task.cancel()

        if self._websocket:
            await self._websocket.close()

    async def _receive_model_event_loop(self, outgoing_queue: Queue) -> None:
        """The loop to receive and handle the model responses.

        Args:
            outgoing_queue (`Queue`):
                The queue to push the model responses to the outside.
        """

        async for message in self._websocket:
            if isinstance(message, bytes):
                message = message.decode("utf-8")

            # Parse the message into ModelEvent instance(s)
            events = await self.parse_api_message(message)

            if events is None:
                continue

            if isinstance(events, ModelEvents.EventBase):
                events = [events]

            for event in events:
                # Send the event to the outgoing queue
                await outgoing_queue.put(event)

    @abstractmethod
    async def parse_api_message(
        self,
        message: str,
    ) -> ModelEvents.EventBase | list[ModelEvents.EventBase] | None:
        """Parse the message received from the realtime model API.

        Args:
            message (`str`):
                The message received from the realtime model API.

        Returns:
            `ModelEvents.EventBase | list[ModelEvents.EventBase] | None`:
                The unified model event(s) in agentscope format.
        """



# -*- coding: utf-8 -*-
"""The Gemini realtime model class."""
import json
from typing import Literal, Any

import shortuuid

from ._events import ModelEvents
from ._base import RealtimeModelBase
from .._logging import logger
from .._utils._common import _get_bytes_from_web_url
from ..message import (
    AudioBlock,
    ImageBlock,
    TextBlock,
    ToolResultBlock,
    ToolUseBlock,
)


class GeminiRealtimeModel(RealtimeModelBase):
    """The Gemini realtime model class."""

    support_input_modalities: list[str] = [
        "audio",
        "text",
        "image",
        "tool_result",
    ]
    """The supported input modalities of the Gemini realtime model."""

    websocket_url: str = (
        "wss://generativelanguage.googleapis.com/ws/"
        "google.ai.generativelanguage.v1alpha.GenerativeService."
        "BidiGenerateContent?key="
    )
    """The websocket URL of the Gemini realtime model API."""

    websocket_headers: dict[str, str] = {
        "Content-Type": "application/json",
    }
    """The websocket headers of the Gemini realtime model API."""

    input_sample_rate: int
    """The input audio sample rate."""

    output_sample_rate: int
    """The output audio sample rate."""

    def __init__(
        self,
        model_name: str,
        api_key: str,
        voice: Literal["Puck", "Charon", "Kore", "Fenrir"] | str = "Puck",
        enable_input_audio_transcription: bool = True,
    ) -> None:
        """Initialize the GeminiRealtimeModel class.

        Args:
            model_name (`str`):
                The model name, e.g. "gemini-2.5-flash-native-audio-preview
                -09-2025". Refers to `official documentation
                <https://ai.google.dev/gemini-api/docs/live?hl=zh-cn&example=mic-stream>`_
                for more details.
            api_key (`str`):
                The Gemini API key for authentication.
            voice (`Literal["Puck", "Charon", "Kore", "Fenrir"] str`,
            defaults to `"Puck"`):
                The voice to be used for text-to-speech.
            enable_input_audio_transcription (`bool`, defaults to `True`):
                Whether to enable input audio transcription.
        """
        super().__init__(model_name)

        self.voice = voice
        self.enable_input_audio_transcription = (
            enable_input_audio_transcription
        )

        # The Gemini realtime API uses 16kHz input and 24kHz output.
        self.input_sample_rate = 16000
        self.output_sample_rate = 24000

        # Set the API key in the websocket URL.
        self.websocket_url = self.websocket_url + api_key

        # Response tracking state.
        # Note: Unlike DashScope/OpenAI which send explicit `response.created`
        # events, Gemini does not. We generate response IDs ourselves using
        # short UUID to ensure uniqueness.
        self._response_id: str | None = None

    def _build_session_config(
        self,
        instructions: str,
        tools: list[dict] | None,
        **kwargs: Any,
    ) -> dict:
        """Build Gemini setup message.

        Gemini Live API requires a "setup" message as the first message
        to configure the session.

        Args:
            instructions (`str`):
                The system instructions for the model.
            tools (`list[dict]`):
                The list of tool JSON schemas.
            **kwargs:
                Additional configuration parameters.

        Returns:
            `dict`:
                The session configuration dict.
        """
        # Model configuration
        session_config: dict = {
            "model": f"models/{self.model_name}",
            "systemInstruction": {
                "parts": [{"text": instructions}],
            },
            "outputAudioTranscription": {},
        }

        # Audio transcription configuration
        if self.enable_input_audio_transcription:
            session_config["inputAudioTranscription"] = {}

        # Generation configuration
        generation_config: dict = {
            "responseModalities": ["AUDIO"],
            **kwargs,
        }

        # Voice configuration
        if self.voice:
            generation_config["speechConfig"] = {
                "voiceConfig": {
                    "prebuiltVoiceConfig": {"voiceName": self.voice},
                },
            }

        session_config["generationConfig"] = generation_config

        # Tools configuration
        if tools:
            session_config["tools"] = self._format_toolkit_schema(tools)

        setup_msg = {"setup": session_config}
        return setup_msg

    def _format_toolkit_schema(
        self,
        schemas: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Format the tools JSON schema into Gemini format.

        Args:
            schemas (`list[dict[str, Any]]`):
                The tool schemas.

        Returns:
            `list[dict[str, Any]]`:
                The formatted tools for Gemini.
        """
        function_declarations = []
        for schema in schemas:
            if "function" not in schema:
                continue
            func = schema["function"].copy()
            function_declarations.append(func)

        return [{"function_declarations": function_declarations}]

    async def send(
        self,
        data: AudioBlock | TextBlock | ImageBlock | ToolResultBlock,
    ) -> None:
        """Send the data to the Gemini realtime model for processing.

        Args:
            data (`AudioBlock | TextBlock | ImageBlock | ToolResultBlock`):
                The data to be sent to the Gemini realtime model.
        """
        from websockets import State

        if not self._websocket or self._websocket.state != State.OPEN:
            raise RuntimeError(
                f"WebSocket is not connected for model {self.model_name}. "
                "Call the `connect` method first.",
            )

        # Type checking
        assert (
            isinstance(data, dict) and "type" in data
        ), "Data must be a dict with a 'type' field."

        # The source must be base64 for audio data
        data_type = data.get("type")

        if data_type not in self.support_input_modalities:
            logger.warning(
                "Gemini Realtime API does not support %s data input. "
                "Supported modalities are: %s",
                data_type,
                ", ".join(self.support_input_modalities),
            )
            return

        # Process the data based on its type
        if data_type == "image":
            to_send_message = await self._parse_image_data(
                ImageBlock(
                    type="image",
                    source=data.get("source"),
                ),
            )

        elif data_type == "audio":
            to_send_message = await self._parse_audio_data(
                AudioBlock(
                    type="audio",
                    source=data.get("source"),
                ),
            )

        elif data_type == "text":
            to_send_message = await self._parse_text_data(
                TextBlock(
                    type="text",
                    text=data.get("text"),
                ),
            )

        elif data_type == "tool_result":
            to_send_message = await self._parse_tool_result_data(
                ToolResultBlock(
                    type="tool_result",
                    name=data.get("name"),
                    output=data.get("output"),
                    id=data.get("id"),
                ),
            )

        else:
            raise RuntimeError(f"Unsupported data type {data_type}")

        if to_send_message:
            await self._websocket.send(to_send_message)

    async def parse_api_message(
        self,
        message: str,
    ) -> ModelEvents.EventBase | list[ModelEvents.EventBase] | None:
        """Parse the message received from the Gemini realtime model API.

        Args:
            message (`str`):
                The message received from the Gemini realtime model API.

        Returns:
            `ModelEvents.EventBase | list[ModelEvents.EventBase] | None`:
                The unified model event(s) in agentscope format.
        """
        try:
            data = json.loads(message)
        except json.decoder.JSONDecodeError:
            return None

        if not isinstance(data, dict):
            return None

        model_event = None

        # ================ Setup related events ================
        if "setupComplete" in data:
            model_event = ModelEvents.ModelSessionCreatedEvent(
                session_id="gemini_session",
            )

        # ================ Server content events ================
        elif "serverContent" in data:
            model_event = await self._parse_server_content(
                data["serverContent"],
            )

        # ================ Tool call events ================
        elif "toolCall" in data:
            model_event = await self._parse_tool_call(data["toolCall"])

        # ================ Tool call cancellation ================
        elif "toolCallCancellation" in data:
            # Tool call was cancelled
            # This effectively ends the current response.
            logger.info(
                "Tool call cancelled: %s",
                data["toolCallCancellation"],
            )
            response_id = self._response_id or ""
            self._response_id = None  # Clear response ID
            model_event = ModelEvents.ModelResponseDoneEvent(
                response_id=response_id,
                input_tokens=0,
                output_tokens=0,
            )

        # ================ Error events ================
        elif "error" in data:
            error = data["error"]
            model_event = ModelEvents.ModelErrorEvent(
                error_type=error.get("status", "unknown"),
                code=str(error.get("code", "unknown")),
                message=error.get("message", "An unknown error occurred."),
            )

        else:
            logger.debug(
                "Unknown Gemini realtime model message keys: %s",
                list(data.keys()),
            )

        return model_event

    def _ensure_response_id(self) -> str:
        """Ensure a response ID exists, creating one if necessary.

        Gemini doesn't send explicit response.created events, so we generate
        the response ID on first audio/text chunk using short UUID.

        Returns:
            `str`: The current response ID.
        """
        if not self._response_id:
            self._response_id = f"resp_{shortuuid.uuid()}"
        # After the check above, _response_id is guaranteed to be non-None
        assert self._response_id is not None
        return self._response_id

    def _parse_model_turn(
        self,
        model_turn: dict,
    ) -> ModelEvents.EventBase | None:
        """Parse the modelTurn content from Gemini API.

        Args:
            model_turn (`dict`):
                The modelTurn dictionary containing parts with audio/text.

        Returns:
            `ModelEvents.EventBase | None`:
                The parsed model event, or None if no valid content found.
        """
        parts = model_turn.get("parts", [])

        for part in parts:
            # Check for audio data
            if "inlineData" in part:
                event = self._parse_inline_data(part["inlineData"])
                if event:
                    return event

            # Check for text data
            if "text" in part:
                text_data = part["text"]
                if text_data:
                    response_id = self._ensure_response_id()
                    return ModelEvents.ModelResponseAudioTranscriptDeltaEvent(
                        response_id=response_id,
                        delta=text_data,
                        item_id="",
                    )

        return None

    def _parse_inline_data(
        self,
        inline_data: dict,
    ) -> ModelEvents.EventBase | None:
        """Parse inline data (audio) from a model turn part.

        Args:
            inline_data (`dict`):
                The inlineData dictionary containing mimeType and data.

        Returns:
            `ModelEvents | None`:
                Audio delta event if valid audio data, None otherwise.
        """
        mime_type = inline_data.get("mimeType", "")
        if not mime_type.startswith("audio/"):
            return None

        audio_data = inline_data.get("data", "")
        if not audio_data:
            return None

        response_id = self._ensure_response_id()
        return ModelEvents.ModelResponseAudioDeltaEvent(
            response_id=response_id,
            item_id="",
            delta=audio_data,
            format={
                "type": "audio/pcm",
                "rate": self.output_sample_rate,
            },
        )

    async def _parse_server_content(
        self,
        server_content: dict,
    ) -> ModelEvents.EventBase | None:
        """Parse the serverContent message from Gemini API.

        Args:
            server_content (`dict`):
                The serverContent dictionary from the API response.

        Returns:
            `ModelEvents.EventBase | None`:
                The unified model event in agentscope format.
        """
        model_event = None

        # Handle model turn (response with audio/text)
        if "modelTurn" in server_content:
            model_event = self._parse_model_turn(server_content["modelTurn"])

        # Handle output transcription
        elif "outputTranscription" in server_content:
            text = server_content["outputTranscription"].get("text", "")
            if text:
                model_event = (
                    ModelEvents.ModelResponseAudioTranscriptDeltaEvent(
                        response_id=self._response_id or "",
                        delta=text,
                        item_id="",
                    )
                )

        # Handle input transcription
        elif "inputTranscription" in server_content:
            text = server_content["inputTranscription"].get("text", "")
            if text:
                model_event = ModelEvents.ModelInputTranscriptionDoneEvent(
                    transcript=text,
                    item_id="",
                )

        # Handle generation complete (response done)
        elif "generationComplete" in server_content:
            response_id = self._response_id or ""
            self._response_id = None
            model_event = ModelEvents.ModelResponseDoneEvent(
                response_id=response_id,
                input_tokens=0,
                output_tokens=0,
            )

        # Handle turn complete
        elif "turnComplete" in server_content:
            logger.debug("Gemini: turnComplete received")
            # turnComplete without generationComplete means interrupted
            if self._response_id:
                response_id = self._response_id
                self._response_id = None
                model_event = ModelEvents.ModelResponseDoneEvent(
                    response_id=response_id,
                    input_tokens=0,
                    output_tokens=0,
                )

        # Handle interrupted
        elif "interrupted" in server_content:
            logger.debug("Gemini: response interrupted")

        return model_event

    async def _parse_tool_call(
        self,
        tool_call: dict,
    ) -> list[ModelEvents.EventBase] | None:
        """Parse the tool call message from Gemini API.

        Args:
            tool_call (`dict`):
                The toolCall dictionary from the API response.

        Returns:
            `list[ModelEvents.EventBase] | None`:
                The unified model event(s) in agentscope format.
        """
        function_calls = tool_call.get("functionCalls", [])
        if not function_calls:
            return None

        events = []
        for func_call in function_calls:
            name = func_call.get("name", "")
            call_id = func_call.get("id", "")
            args = func_call.get("args", {})

            # Return the accumulated arguments instead of just the delta
            model_event = ModelEvents.ModelResponseToolUseDoneEvent(
                response_id=self._response_id or "",
                item_id="",
                tool_use=ToolUseBlock(
                    type="tool_use",
                    id=call_id,
                    name=name,
                    input=args,
                    raw_input=json.dumps(args, ensure_ascii=False),
                ),
            )
            events.append(model_event)

        return events if events else None

    async def _parse_image_data(self, block: ImageBlock) -> str | None:
        """Parse the image data block to the format required by the Gemini
        realtime model API.

        Args:
            block (`ImageBlock`):
                The image data block.

        Returns:
            `str | None`: The parsed message to be sent to the Gemini realtime
            model API.
        """
        source = block.get("source", {})
        source_type = source.get("type", "")
        # media_type is in Base64Source, use default for URLSource
        media_type = source.get("media_type", "image/jpeg")

        if source_type == "base64":
            image_data = source.get("data", "")
        elif source_type == "url":
            image_data = _get_bytes_from_web_url(str(source.get("url", "")))
        else:
            raise ValueError(f"Unsupported image source type: {source_type}")

        return json.dumps(
            {
                "realtimeInput": {
                    "video": {
                        "mimeType": media_type,
                        "data": image_data,
                    },
                },
            },
        )

    async def _parse_audio_data(self, block: AudioBlock) -> str:
        """Parse the audio data block to the format required by the Gemini
        realtime model API.

        Args:
            block (`AudioBlock`):
                The audio data block.

        Returns:
            `str`: The parsed message to be sent to the Gemini realtime
            model API.
        """
        source = block.get("source", {})
        source_type = source.get("type", "")

        if source_type == "base64":
            audio_data = source.get("data", "")
        elif source_type == "url":
            audio_data = _get_bytes_from_web_url(str(source.get("url", "")))
        else:
            raise ValueError(f"Unsupported audio source type: {source_type}")

        return json.dumps(
            {
                "realtimeInput": {
                    "audio": {
                        "mimeType": f"audio/pcm;rate={self.input_sample_rate}",
                        "data": audio_data,
                    },
                },
            },
        )

    async def _parse_text_data(self, block: TextBlock) -> str:
        """Parse the text data block to the format required by the Gemini
        realtime model API.

        Args:
            block (`TextBlock`):
                The text data block.

        Returns:
            `str`: The parsed message to be sent to the Gemini realtime
            model API.
        """
        text = block.get("text", "")

        return json.dumps(
            {
                "clientContent": {
                    "turns": [
                        {
                            "role": "user",
                            "parts": [{"text": text}],
                        },
                    ],
                    # TODO: should be set to False?
                    "turnComplete": True,
                },
            },
        )

    async def _parse_tool_result_data(self, block: ToolResultBlock) -> str:
        """Parse the tool result data block to the format required by the
        Gemini realtime model API.

        Args:
            block (`ToolResultBlock`):
                The tool result data block.

        Returns:
            `str`: The parsed message to be sent to the Gemini realtime
            model API.
        """
        tool_id = block.get("id", "")
        tool_name = block.get("name", "")
        output = block.get("output", "")

        # Extract text from list of blocks (most common case)
        if isinstance(output, list):
            text = "".join(
                str(item.get("text", ""))
                if isinstance(item, dict) and item.get("type") == "text"
                else str(item)
                for item in output
            )
            result_obj = {"result": text}
        elif isinstance(output, str):
            try:
                result_obj = json.loads(output)
            except json.JSONDecodeError:
                result_obj = {"result": output}
        else:
            result_obj = (
                output if isinstance(output, dict) else {"result": str(output)}
            )

        return json.dumps(
            {
                "toolResponse": {
                    "functionResponses": [
                        {
                            "id": tool_id,
                            "name": tool_name,
                            "response": result_obj,
                        },
                    ],
                },
            },
        )



# -*- coding: utf-8 -*-
"""The dashscope realtime model class."""
import json
from typing import Literal, Any

import shortuuid

from ._events import ModelEvents
from ._base import RealtimeModelBase
from .._logging import logger
from .._utils._common import _get_bytes_from_web_url
from ..message import AudioBlock, TextBlock, ImageBlock, ToolResultBlock


class DashScopeRealtimeModel(RealtimeModelBase):
    """The DashScope realtime model class.

    TODO:
     - Support non-VAD mode
     - Support update session config during the session
    """

    support_input_modalities: list[str] = ["text", "audio", "image"]
    """The supported input modalities of the DashScope realtime model."""

    support_tools: bool = False
    """The DashScope Realtime API doesn't support tools yet (last updated in
    20260129)."""

    websocket_url: str = (
        "wss://dashscope.aliyuncs.com/api-ws/v1/realtime?model={model_name}"
    )
    """The websocket URL of the DashScope realtime model API."""

    websocket_headers: dict[str, str]
    """The websocket headers of the DashScope realtime model API."""

    input_sample_rate: int
    """The input audio sample rate."""

    output_sample_rate: int
    """The output audio sample rate."""

    def __init__(
        self,
        model_name: str,
        api_key: str,
        voice: str
        | Literal["Cherry", "Serena", "Ethan", "Chelsie"] = "Cherry",
        enable_input_audio_transcription: bool = True,
    ) -> None:
        """Initialize the DashScopeRealtimeModel class.

        Args:
            model_name (`str`):
                The model name, e.g. "qwen3-omni-flash-realtime".
            api_key (`str`):
                The API key for authentication.
            voice (`str | Literal["Cherry", "Serena", "Ethan", "Chelsie"]`, \
            defaults to `"Cherry"`):
                The voice to be used for text-to-speech.
            enable_input_audio_transcription (`bool`, defaults to `True`):
                Whether to enable input audio transcription.
        """
        super().__init__(model_name)

        self.voice = voice
        self.enable_input_audio_transcription = (
            enable_input_audio_transcription
        )

        # The dashscope realtime API requires 16kHz input sample rate
        # for all models.
        self.input_sample_rate = 16000

        # The output sample rate depends on the model.
        # For "qwen3-omni-flash-realtime" models, it's 24kHz.
        # For others, it's 16kHz.
        if model_name.startswith("qwen3-omni-flash-realtime"):
            self.output_sample_rate = 24000
        else:
            self.output_sample_rate = 16000

        # Set the model name in the websocket URL.
        self.websocket_url = self.websocket_url.format(model_name=model_name)

        # Set the API key in the websocket headers.
        self.websocket_headers = {
            "Authorization": f"Bearer {api_key}",
            "X-DashScope-DataInspection": "disable",
        }

        # Record the response ID for the current session.
        self._response_id = ""

    def _build_session_config(
        self,
        instructions: str,
        tools: list[dict] | None,
        **kwargs: Any,
    ) -> dict:
        """Build the session configuration."""
        session_config: dict = {
            "instructions": instructions,
            # The output modalities of the model
            "modalities": ["audio", "text"],
            "input_audio_format": "pcm" + str(self.input_sample_rate // 1000),
            "output_audio_format": "pcm"
            + str(self.output_sample_rate // 1000),
            "voice": self.voice,
            **kwargs,
        }

        # Input audio transcription
        if self.enable_input_audio_transcription:
            session_config["input_audio_transcription"] = {
                "model": "gummy-realtime-v1",
            }

        # By default, we enable the VAD capability
        # TODO: support none-VAD mode
        session_config["turn_detection"] = {
            "type": "server_vad",
            "threshold": 0.5,
            "silence_duration_ms": 800,
        }

        return {
            "type": "session.update",
            "session": session_config,
        }

    async def send(
        self,
        data: AudioBlock | TextBlock | ImageBlock | ToolResultBlock,
    ) -> None:
        """Send the data to the DashScope realtime model for processing.

        .. note:: The DashScope Realtime API currently only supports audio and
        image data input.

        Args:
            data (`AudioBlock | TextBlock | ImageBlock | ToolResultBlock`):
                The data to be sent to the DashScope realtime model.
        """
        from websockets import State

        if not self._websocket or self._websocket.state != State.OPEN:
            raise RuntimeError(
                f"WebSocket is not connected for model {self.model_name}. "
                "Call the `connect` method first.",
            )

        # Type checking
        assert (
            isinstance(data, dict) and "type" in data
        ), "Data must be a dict with a 'type' field."

        # The source must be base64 for audio data
        data_type = data.get("type")

        if data_type not in self.support_input_modalities:
            logger.warning(
                "DashScope Realtime API does not support %s data input. "
                "Supported modalities are: %s",
                data_type,
                ", ".join(self.support_input_modalities),
            )
            return

        # Process the data based on its type
        if data_type == "image":
            to_send_message = await self._parse_image_data(
                ImageBlock(
                    type="image",
                    source=data.get("source"),
                ),
            )

        elif data_type == "audio":
            to_send_message = await self._parse_audio_data(
                AudioBlock(
                    type="audio",
                    source=data.get("source"),
                ),
            )

        elif data_type == "text":
            # TODO: The following code doesn't work and cannot support text
            #  input yet.
            to_send_message = json.dumps(
                {
                    "event_id": shortuuid.uuid(),
                    "type": "response.create",
                    "response": {
                        "instructions": data.get("text", ""),
                    },
                },
                ensure_ascii=False,
            )

        else:
            raise ValueError(
                f"Unsupported data type: {data_type}",
            )

        await self._websocket.send(to_send_message)

    async def parse_api_message(
        self,
        message: str,
    ) -> ModelEvents.EventBase | list[ModelEvents.EventBase] | None:
        """Parse the message received from the DashScope realtime model API.

        Args:
            message (`str`):
                The message received from the DashScope realtime model API.

        Returns:
            `ModelEvents.EventBase | list[ModelEvents.EventBase] | None`:
                The unified model event(s) in agentscope format.
        """
        try:
            data = json.loads(message)
        except json.decoder.JSONDecodeError:
            return None

        if not isinstance(data, dict):
            return None

        model_event = None
        match data.get("type", ""):
            # ================ Session related events ================
            case "session.created":
                model_event = ModelEvents.ModelSessionCreatedEvent(
                    session_id=data.get("session", {}).get("id", ""),
                )

            case "session.updated":
                # TODO: handle the session updated event
                pass

            # ================ Response related events ================
            case "response.created":
                self._response_id = data.get("response", {}).get("id", "")
                model_event = ModelEvents.ModelResponseCreatedEvent(
                    response_id=self._response_id,
                )

            case "response.done":
                response = data.get("response", {})
                response_id = response.get("id", "") or self._response_id
                usage = response.get("usage", {})
                model_event = ModelEvents.ModelResponseDoneEvent(
                    response_id=response_id,
                    input_tokens=usage.get("input_tokens", 0),
                    output_tokens=usage.get("output_tokens", 0),
                )
                # clear the response id
                self._response_id = ""

            case "response.audio.delta":
                audio_data = data.get("delta", "")
                if audio_data:
                    model_event = ModelEvents.ModelResponseAudioDeltaEvent(
                        response_id=self._response_id,
                        item_id=data.get("item_id", ""),
                        delta=audio_data,
                        format={
                            "type": "audio/pcm",
                            "rate": self.output_sample_rate,
                        },
                    )

            case "response.audio.done":
                model_event = ModelEvents.ModelResponseAudioDoneEvent(
                    response_id=self._response_id,
                    item_id=data.get("item_id", ""),
                )

            # ================ Transcription related events ================

            case "response.audio_transcript.delta":
                transcript_data = data.get("delta", "")
                if transcript_data:
                    model_event = (
                        ModelEvents.ModelResponseAudioTranscriptDeltaEvent(
                            response_id=self._response_id,
                            delta=transcript_data,
                            item_id=data.get("item_id", ""),
                        )
                    )

            case "response.audio_transcript.done":
                model_event = (
                    ModelEvents.ModelResponseAudioTranscriptDoneEvent(
                        response_id=self._response_id,
                        item_id=data.get("item_id", ""),
                    )
                )

            case "conversation.item.input_audio_transcription.completed":
                transcript_data = data.get("transcript", "")
                if transcript_data:
                    model_event = ModelEvents.ModelInputTranscriptionDoneEvent(
                        transcript=transcript_data,
                        item_id=data.get("item_id", ""),
                    )

            # ================= VAD related events =================
            case "input_audio_buffer.speech_started":
                model_event = ModelEvents.ModelInputStartedEvent(
                    item_id=data.get("item_id", ""),
                    audio_start_ms=data.get("audio_start_ms", 0),
                )

            case "input_audio_buffer.speech_stopped":
                model_event = ModelEvents.ModelInputDoneEvent(
                    item_id=data.get("item_id", ""),
                    audio_end_ms=data.get("audio_end_ms", 0),
                )

            # ================= Error events =================
            case "error":
                error = data.get("error", {})
                model_event = ModelEvents.ModelErrorEvent(
                    error_type=error.get("type", "unknown"),
                    code=error.get("code", "unknown"),
                    message=error.get("message", "An unknown error occurred."),
                )

            # ================= Unknown events =================
            case _:
                logger.debug(
                    "Unknown DashScope realtime model event type: %s",
                    data.get("type", None),
                )

        return model_event

    async def _parse_image_data(self, block: ImageBlock) -> str:
        """Parse the image data block to the format required by the DashScope
        realtime model API.

        Args:
            block (`ImageBlock`):
                The image data block.

        Returns:
            `str`: The parsed message to be sent to the DashScope realtime
            model API.
        """
        if block["source"]["type"] == "base64":
            return json.dumps(
                {
                    "type": "input_image_buffer.append",
                    "image": block["source"]["data"],
                },
            )

        if block["source"]["type"] == "url":
            image = _get_bytes_from_web_url(block["source"]["url"])
            return json.dumps(
                {
                    "type": "input_image_url.append",
                    "image_url": image,
                },
            )

        raise ValueError(
            f"Unsupported image source type: {block['source']['type']}",
        )

    async def _parse_audio_data(self, block: AudioBlock) -> str:
        """Parse the audio data block to the format required by the DashScope
        realtime model API.

        Args:
            block (`AudioBlock`):
                The audio data block.

        Returns:
            `str`: The parsed message to be sent to the DashScope realtime
            model API.
        """
        source_type = block["source"]["type"]

        if source_type == "base64":
            audio_data = block["source"]["data"]

        elif source_type == "url":
            audio_data = _get_bytes_from_web_url(block["source"]["url"])

        else:
            raise ValueError(
                f"Unsupported audio source type: {source_type}",
            )

        return json.dumps(
            {
                "type": "input_audio_buffer.append",
                "audio": audio_data,
            },
        )



# -*- coding: utf-8 -*-
"""The TTS model base class."""

from abc import ABC, abstractmethod
from typing import Any, AsyncGenerator

from agentscope.message import Msg

from ._tts_response import TTSResponse


class TTSModelBase(ABC):
    """Base class for TTS models in AgentScope.

    This base class provides general abstraction for both realtime and
    non-realtime TTS models (depending on whether streaming input is
    supported).

    For non-realtime TTS models, the `synthesize` method is used to
    synthesize speech from the input text. You only need to implement the
    `_call_api` method to handle the TTS API calls.

    For realtime TTS models, its lifecycle is managed via the async context
    manager or calling `connect` and `close` methods. The `push` method will
    append text chunks and return the received TTS response, while the
    `synthesize` method will block until the full speech is synthesized.
    You need to implement the `connect`, `close`, and `_call_api` methods
    to handle the TTS API calls and resource management.
    """

    supports_streaming_input: bool = False
    """If the TTS model class supports streaming input."""

    model_name: str
    """The name of the TTS model."""

    stream: bool
    """Whether to use streaming synthesis if supported by the model."""

    def __init__(self, model_name: str, stream: bool) -> None:
        """Initialize the TTS model base class.

        Args:
            model_name (`str`):
                The name of the TTS model
            stream  (`bool`):
                Whether to use streaming synthesis if supported by the model.
        """
        self.model_name = model_name
        self.stream = stream

    async def __aenter__(self) -> "TTSModelBase":
        """Enter the async context manager and initialize resources if
        needed."""
        if self.supports_streaming_input:
            await self.connect()

        return self

    async def __aexit__(
        self,
        exc_type: Any,
        exc_value: Any,
        traceback: Any,
    ) -> None:
        """Exit the async context manager and clean up resources if needed."""
        if self.supports_streaming_input:
            await self.close()

    async def connect(self) -> None:
        """Connect to the TTS model and initialize resources. For non-realtime
        TTS models, leave this method empty.

        .. note:: Only needs to be implemented for realtime TTS models.

        """
        raise NotImplementedError(
            f"The connect method is not implemented for "
            f"{self.__class__.__name__} class.",
        )

    async def close(self) -> None:
        """Close the connection to the TTS model and clean up resources. For
        non-realtime TTS models, leave this method empty.

        .. note:: Only needs to be implemented for realtime TTS models.

        """
        raise NotImplementedError(
            "The close method is not implemented for "
            f"{self.__class__.__name__} class.",
        )

    async def push(
        self,
        msg: Msg,
        **kwargs: Any,
    ) -> TTSResponse:
        """Append text to be synthesized and return the received TTS response.
        Note this method is non-blocking, and maybe return an empty response
        if no audio is received yet.

        To receive all the synthesized speech, call the `synthesize` method
        after pushing all the text chunks.

        .. note:: Only needs to be implemented for realtime TTS models.

        Args:
            msg (`Msg`):
                The message to be synthesized. The `msg.id` identifies the
                streaming input request.
            **kwargs (`Any`):
                Additional keyword arguments to pass to the TTS API call.

        Returns:
            `TTSResponse`:
                The TTSResponse containing audio block.
        """
        raise NotImplementedError(
            "The push method is not implemented for "
            f"{self.__class__.__name__} class.",
        )

    @abstractmethod
    async def synthesize(
        self,
        msg: Msg | None = None,
        **kwargs: Any,
    ) -> TTSResponse | AsyncGenerator[TTSResponse, None]:
        """Synthesize speech from the appended text. Different from the `push`
        method, this method will block until the full speech is synthesized.

        Args:
            msg (`Msg | None`, defaults to `None`):
                The message to be synthesized. If `None`, this method will
                wait for all previously pushed text to be synthesized, and
                return the last synthesized TTSResponse.

        Returns:
            `TTSResponse | AsyncGenerator[TTSResponse, None]`:
                The TTSResponse containing audio blocks, or an async generator
                yielding TTSResponse objects in streaming mode.
        """



# -*- coding: utf-8 -*-
"""DashScope SDK TTS model implementation using MultiModalConversation API."""
from typing import (
    Any,
    Literal,
    AsyncGenerator,
    Generator,
    TYPE_CHECKING,
)

from ._tts_base import TTSModelBase
from ._tts_response import TTSResponse
from ..message import Msg, AudioBlock, Base64Source
from ..types import JSONSerializableObject

if TYPE_CHECKING:
    from dashscope.api_entities.dashscope_response import (
        MultiModalConversationResponse,
    )

else:
    MultiModalConversationResponse = (
        "dashscope.api_entities.dashscope_response."
        "MultiModalConversationResponse"
    )


class DashScopeTTSModel(TTSModelBase):
    """DashScope TTS model implementation using MultiModalConversation API.
    For more details, please see the `official document
    <https://bailian.console.aliyun.com/?tab=doc#/doc/?type=model&url=2879134>`_.
    """

    supports_streaming_input: bool = False
    """Whether the model supports streaming input."""

    def __init__(
        self,
        api_key: str,
        model_name: str = "qwen3-tts-flash",
        voice: Literal["Cherry", "Serena", "Ethan", "Chelsie"]
        | str = "Cherry",
        language_type: str = "Auto",
        stream: bool = True,
        generate_kwargs: dict[str, JSONSerializableObject] | None = None,
    ) -> None:
        """Initialize the DashScope SDK TTS model.

        .. note:: More details about the parameters, such as `model_name`,
        `voice`, and language_type can be found in the `official document
        <https://bailian.console.aliyun.com/?tab=doc#/doc/?type=model&url=2879134>`_.

        Args:
            api_key (`str`):
                The DashScope API key. Required.
            model_name (`str`, defaults to "qwen3-tts-flash"):
                The TTS model name. Supported models are qwen3-tts-flash,
                qwen-tts, etc.
            voice (`Literal["Cherry", "Serena", "Ethan", "Chelsie"] | str`, \
             defaults to "Cherry"):
                The voice to use. Supported voices are "Cherry", "Serena",
                "Ethan", "Chelsie", etc.
            language_type (`str`, default to "Auto"):
                The language type. Should match the text language for
                correct pronunciation and natural intonation.
            generate_kwargs (`dict[str, JSONSerializableObject] | None`, \
             optional):
               The extra keyword arguments used in Dashscope TTS API
               generation, e.g. `temperature`, `seed`.
        """
        super().__init__(model_name=model_name, stream=stream)

        self.api_key = api_key
        self.voice = voice
        self.language_type = language_type
        self.generate_kwargs = generate_kwargs or {}

    async def synthesize(
        self,
        msg: Msg | None = None,
        **kwargs: Any,
    ) -> TTSResponse | AsyncGenerator[TTSResponse, None]:
        """Call the DashScope TTS API to synthesize speech from text.

        Args:
            msg (`Msg | None`, optional):
                The message to be synthesized.
            **kwargs (`Any`):
                Additional keyword arguments to pass to the TTS API call.

        Returns:
            `TTSResponse | AsyncGenerator[TTSResponse, None]`:
                The TTS response or an async generator yielding TTSResponse
                objects in streaming mode.
        """

        if msg is None:
            return TTSResponse(content=None)

        text = msg.get_text_content()

        import dashscope

        # Call DashScope TTS API with streaming mode
        response = dashscope.MultiModalConversation.call(
            model=self.model_name,
            api_key=self.api_key,
            text=text,
            voice=self.voice,
            language_type=self.language_type,
            stream=True,
            **self.generate_kwargs,
            **kwargs,
        )

        if self.stream:
            return self._parse_into_async_generator(response)

        audio_data = ""
        for chunk in response:
            if chunk.output is not None:
                audio_data += chunk.output.audio.data

        res = TTSResponse(
            content=AudioBlock(
                type="audio",
                source=Base64Source(
                    type="base64",
                    data=audio_data,
                    media_type="audio/pcm;rate=24000",
                ),
            ),
        )
        return res

    @staticmethod
    async def _parse_into_async_generator(
        response: Generator[MultiModalConversationResponse, None, None],
    ) -> AsyncGenerator[TTSResponse, None]:
        """Parse the TTS response into an async generator.

        Args:
            response (`Generator[MultiModalConversationResponse, None, None]`):
                The streaming response from DashScope TTS API.

        Returns:
            `AsyncGenerator[TTSResponse, None]`:
                An async generator yielding TTSResponse objects.
        """
        audio_data = ""
        for chunk in response:
            if chunk.output is not None:
                audio = chunk.output.audio
                if audio and audio.data:
                    audio_data += audio.data
                    yield TTSResponse(
                        content=AudioBlock(
                            type="audio",
                            source=Base64Source(
                                type="base64",
                                data=audio_data,
                                media_type="audio/pcm;rate=24000",
                            ),
                        ),
                        is_last=False,
                    )
        yield TTSResponse(
            content=AudioBlock(
                type="audio",
                source=Base64Source(
                    type="base64",
                    data=audio_data,
                    media_type="audio/pcm;rate=24000",
                ),
            ),
            is_last=True,
        )



# -*- coding: utf-8 -*-
# pylint: disable=too-many-branches
"""The dashscope formatter module."""

import json
import os.path
from typing import Any

from ._truncated_formatter_base import TruncatedFormatterBase
from .._logging import logger
from .._utils._common import _is_accessible_local_file
from ..message import (
    Msg,
    TextBlock,
    ImageBlock,
    AudioBlock,
    VideoBlock,
    ToolUseBlock,
    ToolResultBlock,
    URLSource,
)
from ..token import TokenCounterBase


def _format_dashscope_media_block(
    block: ImageBlock | AudioBlock,
) -> dict[str, str]:
    """Format an image or audio block for DashScope API.

    Args:
        block (`ImageBlock` | `AudioBlock`):
            The image or audio block to format.

    Returns:
        `dict[str, str]`:
            A dictionary with "image" or "audio" key and the formatted URL or
            data URI as value.

    Raises:
        `NotImplementedError`:
            If the source type is not supported.
    """
    typ = block["type"]
    source = block["source"]
    if source["type"] == "url":
        url = source["url"]
        if _is_accessible_local_file(url):
            return {typ: "file://" + os.path.abspath(url)}
        else:
            # treat as web url
            return {typ: url}

    elif source["type"] == "base64":
        media_type = source["media_type"]
        base64_data = source["data"]
        return {
            typ: f"data:{media_type};base64,{base64_data}",
        }

    else:
        raise NotImplementedError(
            f"Unsupported source type '{source.get('type')}' "
            f"for {typ} block.",
        )


def _reformat_messages(
    messages: list[dict[str, Any]],
) -> list[dict[str, Any]]:
    """Reformat the content to be compatible with HuggingFaceTokenCounter.

     This function processes a list of messages and converts multi-part
     text content into single string content when all parts are plain text.
     This is necessary for compatibility with HuggingFaceTokenCounter which
     expects simple string content rather than structured content with
     multiple parts.

    Args:
        messages (list[dict[str, Any]]):
            A list of message dictionaries where each message may contain a
            "content" field. The content can be either:
            - A string (unchanged)
            - A list of content items, where each item is a dict that may
                contain "text", "type", and other fields

    Returns:
        list[dict[str, Any]]:
            A list of reformatted messages. For messages where all content
            items are plain text (have "text" field and either no "type"
            field or "type" == "text"), the content list is converted to a
            single newline-joined string. Other messages remain unchanged.

    Example:
        .. code-block:: python

            # Case 1: All text content - will be converted
            messages = [
                {
                    "role": "user",
                    "content": [
                        {"text": "Hello", "type": "text"},
                        {"text": "World", "type": "text"}
                    ]
                }
            ]
            result = _reformat_messages(messages)
            print(result[0]["content"])
            # Output: "Hello\nWorld"

            # Case 2: Mixed content - will remain unchanged
            messages = [
                {
                    "role": "user",
                    "content": [
                        {"text": "Hello", "type": "text"},
                        {"image_url": "...", "type": "image"}
                    ]
                }
            ]

            result = _reformat_messages(messages)  # remain unchanged
            print(type(result[0]["content"]))
            # Output: <class 'list'>

    """
    for message in messages:
        content = message.get("content", [])

        is_all_text = True
        texts = []
        for item in content:
            if not isinstance(item, dict) or "text" not in item:
                is_all_text = False
                break
            if "type" in item and item["type"] != "text":
                is_all_text = False
                break
            if item["text"]:
                texts.append(item["text"])

        if is_all_text and texts:
            message["content"] = "\n".join(texts)

    return messages


class DashScopeChatFormatter(TruncatedFormatterBase):
    """The DashScope formatter class for chatbot scenario, where only a user
    and an agent are involved. We use the `role` field to identify different
    entities in the conversation.

    .. warning::
        Known Issues with DashScope API:

        1. **Missing content field**: When messages lack the 'content' field,
           qwen-vl-max models will raise ``KeyError: 'content'``.

        2. **None content value**: When content is ``None``, qwen-vl-max models
           will raise ``TypeError: 'NoneType' object is not iterable``.

        3. **Empty text in content**: When content contains
           ``[{"text": None}]``, qwen3-max may repeatedly invoke tools
           multiple times. Note that when qwen3-max initiates tool calls,
           the returned message contains ``"content": ""``.

        To avoid these issues, this formatter assigns content as an empty
        list ``[]`` for messages without valid content blocks.

    """

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = False
    """Whether support multi-agent conversations"""

    support_vision: bool = True
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        ImageBlock,
        AudioBlock,
        VideoBlock,
        ToolUseBlock,
        ToolResultBlock,
    ]

    def __init__(
        self,
        promote_tool_result_images: bool = False,
        promote_tool_result_audios: bool = False,
        promote_tool_result_videos: bool = False,
        token_counter: TokenCounterBase | None = None,
        max_tokens: int | None = None,
    ) -> None:
        """Initialize the DashScope chat formatter.

        Args:
            promote_tool_result_images (`bool`, defaults to `False`):
                Whether to promote images from tool results to user messages.
                Most LLM APIs don't support images in tool result blocks, but
                do support them in user message blocks. When `True`, images are
                extracted and appended as a separate user message with
                explanatory text indicating their source.
            promote_tool_result_audios (`bool`, defaults to `False`):
                Whether to promote audios from tool results to user messages.
                Most LLM APIs don't support audios in tool result blocks, but
                do support them in user message blocks. When `True`, audios are
                extracted and appended as a separate user message with
                explanatory text indicating their source.
            promote_tool_result_videos (`bool`, defaults to `False`):
                Whether to promote videos from tool results to user messages.
                Most LLM APIs don't support videos in tool result blocks, but
                do support them in user message blocks. When `True`, videos are
                extracted and appended as a separate user message with
                explanatory text indicating their source.
            token_counter (`TokenCounterBase | None`, optional):
                A token counter instance used to count tokens in the messages.
                If not provided, the formatter will format the messages
                without considering token limits.
            max_tokens (`int | None`, optional):
                The maximum number of tokens allowed in the formatted
                messages. If not provided, the formatter will not truncate
                the messages.
        """
        super().__init__(token_counter, max_tokens)
        self.promote_tool_result_images = promote_tool_result_images
        self.promote_tool_result_audios = promote_tool_result_audios
        self.promote_tool_result_videos = promote_tool_result_videos

    async def _format(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Format message objects into DashScope API format.

        Args:
            msgs (`list[Msg]`):
                The list of message objects to format.

        Returns:
            `list[dict[str, Any]]`:
                The formatted messages as a list of dictionaries.
        """
        self.assert_list_of_msgs(msgs)

        formatted_msgs: list[dict] = []

        i = 0
        while i < len(msgs):
            msg = msgs[i]
            content_blocks: list[dict[str, Any]] = []
            tool_calls = []

            for block in msg.get_content_blocks():
                typ = block.get("type")

                if typ == "text":
                    content_blocks.append(
                        {
                            "text": block.get("text"),
                        },
                    )

                elif typ in ["image", "audio", "video"]:
                    content_blocks.append(
                        _format_dashscope_media_block(
                            block,  # type: ignore[arg-type]
                        ),
                    )

                elif typ == "tool_use":
                    tool_calls.append(
                        {
                            "id": block.get("id"),
                            "type": "function",
                            "function": {
                                "name": block.get("name"),
                                "arguments": json.dumps(
                                    block.get("input", {}),
                                    ensure_ascii=False,
                                ),
                            },
                        },
                    )

                elif typ == "tool_result":
                    (
                        textual_output,
                        multimodal_data,
                    ) = self.convert_tool_result_to_string(block["output"])

                    # First add the tool result message in DashScope API format
                    formatted_msgs.append(
                        {
                            "role": "tool",
                            "tool_call_id": block.get("id"),
                            "content": textual_output,
                            "name": block.get("name"),
                        },
                    )

                    # Then, handle the multimodal data if any
                    promoted_blocks: list = []
                    for url, multimodal_block in multimodal_data:
                        if (
                            multimodal_block["type"] == "image"
                            and self.promote_tool_result_images
                        ):
                            promoted_blocks.extend(
                                [
                                    TextBlock(
                                        type="text",
                                        text=f"\n- The image from '{url}': ",
                                    ),
                                    ImageBlock(
                                        type="image",
                                        source=URLSource(
                                            type="url",
                                            url=url,
                                        ),
                                    ),
                                ],
                            )
                        elif (
                            multimodal_block["type"] == "audio"
                            and self.promote_tool_result_audios
                        ):
                            promoted_blocks.extend(
                                [
                                    TextBlock(
                                        type="text",
                                        text=f"\n- The audio from '{url}': ",
                                    ),
                                    AudioBlock(
                                        type="audio",
                                        source=URLSource(
                                            type="url",
                                            url=url,
                                        ),
                                    ),
                                ],
                            )
                        elif (
                            multimodal_block["type"] == "video"
                            and self.promote_tool_result_videos
                        ):
                            promoted_blocks.extend(
                                [
                                    TextBlock(
                                        type="text",
                                        text=f"\n- The video from '{url}': ",
                                    ),
                                    VideoBlock(
                                        type="video",
                                        source=URLSource(
                                            type="url",
                                            url=url,
                                        ),
                                    ),
                                ],
                            )

                    if promoted_blocks:
                        # Insert promoted blocks as new user message(s)
                        promoted_blocks = [
                            TextBlock(
                                type="text",
                                text="<system-info>The following are "
                                f"the media contents from the tool "
                                f"result of '{block['name']}':",
                            ),
                            *promoted_blocks,
                            TextBlock(
                                type="text",
                                text="</system-info>",
                            ),
                        ]

                        msgs.insert(
                            i + 1,
                            Msg(
                                name="user",
                                content=promoted_blocks,
                                role="user",
                            ),
                        )

                else:
                    logger.warning(
                        "Unsupported block type %s in the message, skipped.",
                        typ,
                    )

            msg_dashscope = {
                "role": msg.role,
                "content": content_blocks,
            }

            if tool_calls:
                msg_dashscope["tool_calls"] = tool_calls

            if msg_dashscope["content"] or msg_dashscope.get("tool_calls"):
                formatted_msgs.append(msg_dashscope)

            # Move to next message
            i += 1

        return _reformat_messages(formatted_msgs)


class DashScopeMultiAgentFormatter(TruncatedFormatterBase):
    """DashScope formatter for multi-agent conversations, where more than
    a user and an agent are involved.

    .. note:: This formatter will combine previous messages (except tool
     calls/results) into a history section in the first system message with
     the conversation history prompt.

    .. note:: For tool calls/results, they will be presented as separate
     messages as required by the DashScope API. Therefore, the tool calls/
     results messages are expected to be placed at the end of the input
     messages.

    .. tip:: Telling the assistant's name in the system prompt is very
     important in multi-agent conversations. So that LLM can know who it
     is playing as.

    """

    support_tools_api: bool = True
    """Whether support tools API"""

    support_multiagent: bool = True
    """Whether support multi-agent conversations"""

    support_vision: bool = True
    """Whether support vision data"""

    supported_blocks: list[type] = [
        TextBlock,
        # Multimodal
        ImageBlock,
        AudioBlock,
        VideoBlock,
        # Tool use
        ToolUseBlock,
        ToolResultBlock,
    ]
    """The list of supported message blocks"""

    def __init__(
        self,
        conversation_history_prompt: str = (
            "# Conversation History\n"
            "The content between <history></history> tags contains "
            "your conversation history\n"
        ),
        promote_tool_result_images: bool = False,
        promote_tool_result_audios: bool = False,
        promote_tool_result_videos: bool = False,
        token_counter: TokenCounterBase | None = None,
        max_tokens: int | None = None,
    ) -> None:
        """Initialize the DashScope multi-agent formatter.

        Args:
            conversation_history_prompt (`str`):
                The prompt to use for the conversation history section.
            promote_tool_result_images (`bool`, defaults to `False`):
                Whether to promote images from tool results to user messages.
                Most LLM APIs don't support images in tool result blocks, but
                do support them in user message blocks. When `True`, images are
                extracted and appended as a separate user message with
                explanatory text indicating their source.
            promote_tool_result_audios (`bool`, defaults to `False`):
                Whether to promote audios from tool results to user messages.
                Most LLM APIs don't support audios in tool result blocks, but
                do support them in user message blocks. When `True`, audios are
                extracted and appended as a separate user message with
                explanatory text indicating their source.
            promote_tool_result_videos (`bool`, defaults to `False`):
                Whether to promote videos from tool results to user messages.
                Most LLM APIs don't support videos in tool result blocks, but
                do support them in user message blocks. When `True`, videos are
                extracted and appended as a separate user message with
                explanatory text indicating their source.
            token_counter (`TokenCounterBase | None`, optional):
                The token counter used for truncation.
            max_tokens (`int | None`, optional):
                The maximum number of tokens allowed in the formatted
                messages. If `None`, no truncation will be applied.
        """
        super().__init__(token_counter=token_counter, max_tokens=max_tokens)
        self.conversation_history_prompt = conversation_history_prompt
        self.promote_tool_result_images = promote_tool_result_images
        self.promote_tool_result_audios = promote_tool_result_audios
        self.promote_tool_result_videos = promote_tool_result_videos

    async def _format_tool_sequence(
        self,
        msgs: list[Msg],
    ) -> list[dict[str, Any]]:
        """Given a sequence of tool call/result messages, format them into
        the required format for the DashScope API.

        Args:
            msgs (`list[Msg]`):
                The list of messages containing tool calls/results to format.

        Returns:
            `list[dict[str, Any]]`:
                A list of dictionaries formatted for the DashScope API.
        """
        return await DashScopeChatFormatter(
            promote_tool_result_images=self.promote_tool_result_images,
            promote_tool_result_audios=self.promote_tool_result_audios,
            promote_tool_result_videos=self.promote_tool_result_videos,
        ).format(msgs)

    async def _format_agent_message(
        self,
        msgs: list[Msg],
        is_first: bool = True,
    ) -> list[dict[str, Any]]:
        """Given a sequence of messages without tool calls/results, format
        them into a user message with conversation history tags. For the
        first agent message, it will include the conversation history prompt.

        Args:
            msgs (`list[Msg]`):
                A list of Msg objects to be formatted.
            is_first (`bool`, defaults to `True`):
                Whether this is the first agent message in the conversation.
                If `True`, the conversation history prompt will be included.

        Returns:
            `list[dict[str, Any]]`:
                A list of dictionaries formatted for the DashScope API.
        """

        if is_first:
            conversation_history_prompt = self.conversation_history_prompt
        else:
            conversation_history_prompt = ""

        # Format into required DashScope format
        formatted_msgs: list[dict] = []

        # Collect the multimodal files
        conversation_blocks = []
        accumulated_text = []
        for msg in msgs:
            for block in msg.get_content_blocks():
                if block["type"] == "text":
                    accumulated_text.append(f"{msg.name}: {block['text']}")

                elif block["type"] in ["image", "audio", "video"]:
                    # Handle the accumulated text as a single block
                    if accumulated_text:
                        conversation_blocks.append(
                            {"text": "\n".join(accumulated_text)},
                        )
                        accumulated_text.clear()

                    if block["source"]["type"] == "url":
                        url = block["source"]["url"]
                        if _is_accessible_local_file(url):
                            conversation_blocks.append(
                                {
                                    block["type"]: "file://"
                                    + os.path.abspath(url),
                                },
                            )
                        else:
                            conversation_blocks.append({block["type"]: url})

                    elif block["source"]["type"] == "base64":
                        media_type = block["source"]["media_type"]
                        base64_data = block["source"]["data"]
                        conversation_blocks.append(
                            {
                                block[
                                    "type"
                                ]: f"data:{media_type};base64,{base64_data}",
                            },
                        )

                    else:
                        logger.warning(
                            "Unsupported block type %s in the message, "
                            "skipped.",
                            block["type"],
                        )

        if accumulated_text:
            conversation_blocks.append({"text": "\n".join(accumulated_text)})

        if conversation_blocks:
            if conversation_blocks[0].get("text"):
                conversation_blocks[0]["text"] = (
                    conversation_history_prompt
                    + "<history>\n"
                    + conversation_blocks[0]["text"]
                )

            else:
                conversation_blocks.insert(
                    0,
                    {
                        "text": conversation_history_prompt + "<history>\n",
                    },
                )

            if conversation_blocks[-1].get("text"):
                conversation_blocks[-1]["text"] += "\n</history>"

            else:
                conversation_blocks.append({"text": "</history>"})

            formatted_msgs.append(
                {
                    "role": "user",
                    "content": conversation_blocks,
                },
            )

        return _reformat_messages(formatted_msgs)

    async def _format_system_message(
        self,
        msg: Msg,
    ) -> dict[str, Any]:
        """Format system message for DashScope API."""
        return {
            "role": "system",
            "content": msg.get_text_content(),
        }



# -*- coding: utf-8 -*-
"""The websocket events generated from the realtime agent and backend."""
from enum import Enum
from typing import Literal

from pydantic import BaseModel

from ._utils import AudioFormat
from ._model_event import ModelEvents
from ...message import ToolUseBlock, ToolResultBlock


class ServerEventType(str, Enum):
    """Types of agent events for backend-to-web communication."""

    # Session lifecycle
    SERVER_SESSION_CREATED = "server_session_created"
    """The session between the web frontend and backend is created."""

    SERVER_SESSION_UPDATED = "server_session_updated"
    """The session between the web frontend and backend is updated."""

    SERVER_SESSION_ENDED = "server_session_ended"
    """The session between the web frontend and backend is ended."""

    # ============== AGENT LIFECYCLE EVENTS ================

    AGENT_READY = "agent_ready"
    """The agent is created and ready to receive inputs."""

    AGENT_ENDED = "agent_ended"
    """The agent ended."""

    # ============== AGENT RESPONSE EVENTS =================

    # Response events
    AGENT_RESPONSE_CREATED = "agent_response_created"
    """The agent starts generating a response."""

    AGENT_RESPONSE_DONE = "agent_response_done"
    """The agent finished generating a response."""

    # ============== Response content events =================

    AGENT_RESPONSE_AUDIO_DELTA = "agent_response_audio_delta"
    """The agent's response audio data delta."""

    AGENT_RESPONSE_AUDIO_DONE = "agent_response_audio_done"
    """The agent's response audio data is complete."""

    AGENT_RESPONSE_AUDIO_TRANSCRIPT_DELTA = (
        "agent_response_audio_transcript_delta"
    )
    """The agent's response audio transcription delta."""

    AGENT_RESPONSE_AUDIO_TRANSCRIPT_DONE = (
        "agent_response_audio_transcript_done"
    )
    """The agent's response audio transcription is complete."""

    AGENT_RESPONSE_TOOL_USE_DELTA = "agent_response_tool_use_delta"
    """The agent's response tool use data delta."""

    AGENT_RESPONSE_TOOL_USE_DONE = "agent_response_tool_use_done"
    """The agent's response tool use data is complete."""

    AGENT_RESPONSE_TOOL_RESULT = "agent_response_tool_result"
    """The tool execution result."""

    # ============== INPUT AUDIO TRANSCRIPTION EVENTS =================

    AGENT_INPUT_TRANSCRIPTION_DELTA = "agent_input_transcription_delta"
    """The input audio transcription delta."""

    AGENT_INPUT_TRANSCRIPTION_DONE = "agent_input_transcription_done"
    """The input audio transcription is complete."""

    # Input detection
    AGENT_INPUT_STARTED = "agent_input_started"
    """Detected the start of user input audio."""

    AGENT_INPUT_DONE = "agent_input_done"
    """Detected the end of user input audio."""

    # ============== ERROR EVENTS =================

    AGENT_ERROR = "agent_error"
    """An error occurred in the backend or agent."""


class ServerEvents:
    """Realtime server events."""

    class EventBase(BaseModel):
        """The base class for all server events, used to unify the type
        hinting."""

    class ServerSessionCreatedEvent(EventBase):
        """Server session created event in the backend"""

        session_id: str
        """The session ID."""

        type: Literal[
            ServerEventType.SERVER_SESSION_CREATED
        ] = ServerEventType.SERVER_SESSION_CREATED
        """The event type."""

    class ServerSessionUpdatedEvent(EventBase):
        """Server session updated event in the backend"""

        session_id: str
        """The session ID."""

        type: Literal[
            ServerEventType.SERVER_SESSION_UPDATED
        ] = ServerEventType.SERVER_SESSION_UPDATED
        """The event type."""

    class ServerSessionEndedEvent(EventBase):
        """Server Session ended event in the backend"""

        session_id: str
        """The session ID."""

        type: Literal[
            ServerEventType.SERVER_SESSION_ENDED
        ] = ServerEventType.SERVER_SESSION_ENDED
        """The event type."""

    class AgentReadyEvent(EventBase):
        """Agent ready event in the backend"""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_READY
        ] = ServerEventType.AGENT_READY
        """The event type."""

    class AgentEndedEvent(EventBase):
        """Agent ended event in the backend"""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_ENDED
        ] = ServerEventType.AGENT_ENDED
        """The event type."""

    class AgentResponseCreatedEvent(EventBase):
        """Response created event in the backend"""

        response_id: str
        """The response ID."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_CREATED
        ] = ServerEventType.AGENT_RESPONSE_CREATED
        """The event type."""

    class AgentResponseDoneEvent(EventBase):
        """Response done event in the backend"""

        response_id: str
        """The response ID."""

        input_tokens: int
        """The number of input tokens used."""

        output_tokens: int
        """The number of output tokens used."""

        metadata: dict[str, str] = {}
        """Additional metadata about the response."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_DONE
        ] = ServerEventType.AGENT_RESPONSE_DONE
        """The event type."""

    class AgentResponseAudioDeltaEvent(EventBase):
        """Response audio delta event in the backend"""

        response_id: str
        """The response ID."""

        item_id: str
        """The response item ID."""

        delta: str
        """The audio chunk data, encoded as base64 string."""

        format: AudioFormat
        """The audio format information."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_AUDIO_DELTA
        ] = ServerEventType.AGENT_RESPONSE_AUDIO_DELTA
        """The event type."""

    class AgentResponseAudioDoneEvent(EventBase):
        """Response audio done event in the backend"""

        response_id: str
        """The response ID."""

        item_id: str
        """The response item ID."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_AUDIO_DONE
        ] = ServerEventType.AGENT_RESPONSE_AUDIO_DONE

    class AgentResponseAudioTranscriptDeltaEvent(EventBase):
        """Response audio transcript delta event in the backend"""

        response_id: str
        """The response ID."""

        item_id: str
        """The response item ID."""

        delta: str
        """The transcript chunk data."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_AUDIO_TRANSCRIPT_DELTA
        ] = ServerEventType.AGENT_RESPONSE_AUDIO_TRANSCRIPT_DELTA
        """The event type."""

    class AgentResponseAudioTranscriptDoneEvent(EventBase):
        """Response audio transcript done event in the backend"""

        response_id: str
        """The response ID."""

        item_id: str
        """The response item ID."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_AUDIO_TRANSCRIPT_DONE
        ] = ServerEventType.AGENT_RESPONSE_AUDIO_TRANSCRIPT_DONE
        """The event type."""

    class AgentResponseToolUseDeltaEvent(EventBase):
        """Response tool use delta event in the backend"""

        response_id: str
        """The response ID."""

        item_id: str
        """The response item ID."""

        tool_use: ToolUseBlock
        """The tool use block delta, the arguments are accumulated in the
        `raw_input` field."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_TOOL_USE_DELTA
        ] = ServerEventType.AGENT_RESPONSE_TOOL_USE_DELTA
        """The event type."""

    class AgentResponseToolUseDoneEvent(EventBase):
        """Response tool use done event in the backend"""

        response_id: str
        """The response ID."""

        item_id: str
        """The response item ID."""

        tool_use: ToolUseBlock
        """The complete tool use block."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_TOOL_USE_DONE
        ] = ServerEventType.AGENT_RESPONSE_TOOL_USE_DONE
        """The event type."""

    class AgentResponseToolResultEvent(EventBase):
        """Response tool result event"""

        tool_result: ToolResultBlock
        """The tool result block."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_RESPONSE_TOOL_RESULT
        ] = ServerEventType.AGENT_RESPONSE_TOOL_RESULT
        """The event type."""

    class AgentInputTranscriptionDeltaEvent(EventBase):
        """Input transcription delta event in the backend"""

        item_id: str
        """The conversation item ID."""

        delta: str
        """The transcription chunk data."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_INPUT_TRANSCRIPTION_DELTA
        ] = ServerEventType.AGENT_INPUT_TRANSCRIPTION_DELTA
        """The event type."""

    class AgentInputTranscriptionDoneEvent(EventBase):
        """Input transcription done event in the backend"""

        transcript: str
        """The complete transcription text."""

        item_id: str
        """The conversation item ID."""

        input_tokens: int | None = None
        """The number of input tokens."""

        output_tokens: int | None = None
        """The number of output tokens."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_INPUT_TRANSCRIPTION_DONE
        ] = ServerEventType.AGENT_INPUT_TRANSCRIPTION_DONE
        """The event type."""

    class AgentInputStartedEvent(EventBase):
        """Input started event in the backend"""

        item_id: str
        """The conversation item ID."""

        audio_start_ms: int
        """The audio start time in milliseconds."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_INPUT_STARTED
        ] = ServerEventType.AGENT_INPUT_STARTED
        """The event type."""

    class AgentInputDoneEvent(EventBase):
        """Input done event in the backend"""

        item_id: str
        """The conversation item ID."""

        audio_end_ms: int
        """The audio end time in milliseconds."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_INPUT_DONE
        ] = ServerEventType.AGENT_INPUT_DONE
        """The event type."""

    class AgentErrorEvent(EventBase):
        """Error event in the backend"""

        error_type: str
        """The error type."""

        code: str
        """The error code."""

        message: str
        """The error message."""

        agent_id: str
        """The agent ID."""

        agent_name: str
        """The agent name."""

        type: Literal[
            ServerEventType.AGENT_ERROR
        ] = ServerEventType.AGENT_ERROR
        """The event type."""

    @classmethod
    def from_model_event(
        cls,
        model_event: ModelEvents.ModelResponseCreatedEvent
        | ModelEvents.ModelResponseDoneEvent
        | ModelEvents.ModelResponseAudioDeltaEvent
        | ModelEvents.ModelResponseAudioDoneEvent
        | ModelEvents.ModelResponseAudioTranscriptDeltaEvent
        | ModelEvents.ModelResponseAudioTranscriptDoneEvent
        | ModelEvents.ModelResponseToolUseDeltaEvent
        | ModelEvents.ModelResponseToolUseDoneEvent
        | ModelEvents.ModelInputTranscriptionDeltaEvent
        | ModelEvents.ModelInputTranscriptionDoneEvent
        | ModelEvents.ModelInputStartedEvent
        | ModelEvents.ModelInputDoneEvent
        | ModelEvents.ModelErrorEvent,
        agent_id: str,
        agent_name: str,
    ) -> EventBase:
        """Convert a model event to a server event quickly with
        1) replace the "model_" prefix with "agent_" in the type field; 2) add
        agent_id and agent_name fields

        Args:
            model_event (`ModelEvents.EventBase`):
                The model event to convert.
            agent_id (`str`):
                The agent ID.
            agent_name (`str`):
                The agent name.

        Returns:
            `ServerEvents.EventBase`:
                The converted server event.
        """
        # Obtain the corresponding agent event class
        cls_name = model_event.__class__.__name__.replace("Model", "Agent")
        agent_event_cls = getattr(cls, cls_name)

        # The data dict of the model event
        model_event_dict = model_event.model_dump()

        # 1) Replace the "model_" prefix with "agent_" in the type field
        if "type" in model_event_dict:
            model_event_dict["type"] = model_event_dict["type"].replace(
                "model_",
                "agent_",
            )

        try:
            # 2) Add agent_id and agent_name fields
            model_event_dict["agent_id"] = agent_id
            model_event_dict["agent_name"] = agent_name
            agent_event = agent_event_cls.model_validate(model_event_dict)

        except Exception as e:
            raise RuntimeError(
                f"Failed to convert model event {model_event} to agent "
                f"event {agent_event_cls}: {e}",
            ) from e

        return agent_event



# -*- coding: utf-8 -*-
"""The JSON file based A2A agent card resolver."""
import json
from pathlib import Path
from typing import TYPE_CHECKING

from ._base import AgentCardResolverBase

if TYPE_CHECKING:
    from a2a.types import AgentCard
else:
    AgentCard = "a2a.types.AgentCard"


class FileAgentCardResolver(AgentCardResolverBase):
    """Agent card resolver that loads AgentCard from a JSON file.

    The JSON file should contain an AgentCard object with the following
    required fields:

    - name (str): The name of the agent
    - url (str): The URL of the agent
    - version (str): The version of the agent
    - capabilities (dict): The capabilities of the agent
    - default_input_modes (list[str]): Default input modes
    - default_output_modes (list[str]): Default output modes
    - skills (list): List of agent skills

    Example JSON file content:

        .. code-block:: json

            {
                "name": "RemoteAgent",
                "url": "http://localhost:8000",
                "description": "A remote A2A agent",
                "version": "1.0.0",
                "capabilities": {},
                "default_input_modes": ["text/plain"],
                "default_output_modes": ["text/plain"],
                "skills": []
            }

    """

    def __init__(
        self,
        file_path: str,
    ) -> None:
        """Initialize the FileAgentCardResolver with the path to the JSON file.

        Args:
            file_path (`str`):
                The path to the JSON file containing the agent card.
        """
        self._file_path = file_path

    async def get_agent_card(self) -> AgentCard:
        """Get the agent card from the JSON file.

        Returns:
            `AgentCard`:
                The agent card loaded from the file.
        """
        from a2a.types import AgentCard

        path = Path(self._file_path)
        if not path.exists():
            raise FileNotFoundError(
                f"Agent card file not found: {self._file_path}",
            )

        if not path.is_file():
            raise ValueError(f"Path is not a file: {self._file_path}")

        with path.open("r", encoding="utf-8") as f:
            agent_json_data = json.load(f)
            return AgentCard.model_validate(agent_json_data)



# -*- coding: utf-8 -*-
"""The Voice chat room"""
import asyncio
from asyncio import Queue

from ..agent import RealtimeAgent
from ..realtime import ClientEvents, ServerEvents


class ChatRoom:
    """The chat room abstraction to broadcast messages among multiple realtime
    agents, and handle the messages from the frontend.
    """

    def __init__(self, agents: list[RealtimeAgent]) -> None:
        """Initialize the ChatRoom class.

        Args:
            agents (`list[RealtimeAgent]`):
                The list of agents participating in the chat room.
        """
        self.agents = agents

        # The queue used to gather messages from all agents and push them to
        # the frontend.
        self._queue = Queue()

        self._task = None

    async def start(self, outgoing_queue: Queue) -> None:
        """Establish connections for all agents in the chat room.

        Args:
            outgoing_queue (`Queue`):
                The queue to push messages to the frontend, which will be used
                by all agents to push their messages.
        """

        for agent in self.agents:
            await agent.start(self._queue)

        # Start the forwarding loop.
        self._task = asyncio.create_task(self._forward_loop(outgoing_queue))

    async def _forward_loop(self, outgoing_queue: Queue) -> None:
        """The loop to forward messages from all agents to the frontend and
        the other agents.

        Args:
            outgoing_queue (`Queue`):
                The queue to push messages to the frontend.
        """

        while True:
            # Obtain the message from the client frontend
            event = await self._queue.get()

            # Only push ServerEvents to the frontend, not ClientEvents
            # to avoid echoing client messages back
            if isinstance(event, ClientEvents.EventBase):
                # Push the message to the frontend queue.
                for agent in self.agents:
                    await agent.handle_input(event)

            elif isinstance(event, ServerEvents.EventBase):
                # Broadcast the message to all agents except the sender.
                # Use create_task instead of gather to avoid blocking

                # Forward the agent/server events to the frontend
                await outgoing_queue.put(event)

                # Broadcast to other agents
                sender_id = getattr(event, "agent_id", None)
                if sender_id:
                    for agent in self.agents:
                        if agent.id != sender_id:
                            await agent.handle_input(event)

    async def stop(self) -> None:
        """Close connections for all agents in the chat room."""

        for agent in self.agents:
            await agent.stop()

        # Close the forwarding loop.
        if not self._task.done():
            self._task.cancel()

    async def handle_input(self, event: ClientEvents.EventBase) -> None:
        """Handle input message from the frontend and distribute it to all
        agents in the chat room.

        Args:
            event (`ClientEvents.EventBase`):
                The event from the frontend.
        """
        await self._queue.put(event)



# -*- coding: utf-8 -*-
"""The OpenAI realtime model class."""
import json
from typing import Literal, Any

from ._events import ModelEvents
from ._base import RealtimeModelBase
from .._logging import logger
from .._utils._common import _get_bytes_from_web_url, _json_loads_with_repair
from ..message import (
    AudioBlock,
    TextBlock,
    ImageBlock,
    ToolResultBlock,
    ToolUseBlock,
)


class OpenAIRealtimeModel(RealtimeModelBase):
    """The OpenAI realtime model class."""

    support_input_modalities: list[str] = ["audio", "text", "tool_result"]
    """The supported input modalities of the OpenAI realtime model."""

    support_tools: bool = True
    """The OpenAI realtime model supports tools API."""

    websocket_url: str = "wss://api.openai.com/v1/realtime?model={model_name}"
    """The websocket URL of the OpenAI realtime model API."""

    websocket_headers: dict[str, str]
    """The websocket headers of the OpenAI realtime model API."""

    input_sample_rate: int
    """The input audio sample rate."""

    output_sample_rate: int
    """The output audio sample rate."""

    def __init__(
        self,
        model_name: str,
        api_key: str,
        voice: Literal["alloy", "echo", "marin", "cedar"] | str = "alloy",
        enable_input_audio_transcription: bool = True,
    ) -> None:
        """Initialize the OpenAIRealtimeModel class.

        Args:
            model_name (`str`):
                The model name, e.g. "gpt-4o-realtime-preview".
            api_key (`str`):
                The API key for authentication.
            voice (`Literal["alloy", "echo", "marin", "cedar"] | str`, \
            defaults to `"alloy"`):
                The voice to be used for text-to-speech.
            enable_input_audio_transcription (`bool`, defaults to `True`):
                Whether to enable input audio transcription.
        """
        super().__init__(model_name)

        self.voice = voice
        self.enable_input_audio_transcription = (
            enable_input_audio_transcription
        )

        # The OpenAI realtime API uses 24kHz for both input and output.
        self.input_sample_rate = 24000
        self.output_sample_rate = 24000

        # Set the model name in the websocket URL.
        self.websocket_url = self.websocket_url.format(model_name=model_name)

        # Set the API key in the websocket headers.
        self.websocket_headers = {
            "Authorization": f"Bearer {api_key}",
            "OpenAI-Beta": "realtime=v1",
        }

        # Record the response ID for the current session.
        self._response_id = ""

        # Tool arguments accumulator for tracking tool call parameters
        self._tool_args_accumulator: dict[str, str] = {}

    def _build_session_config(
        self,
        instructions: str,
        tools: list[dict] | None,
        **kwargs: Any,
    ) -> dict:
        """Build the session configuration for the OpenAI realtime model."""

        session_config: dict[str, Any] = {
            "type": "realtime",
            "output_modalities": ["audio"],
            "audio": {
                "input": {
                    "turn_detection": {
                        "type": "server_vad",
                        "create_response": True,
                    },
                },
                "output": {
                    "voice": self.voice,
                },
            },
            "instructions": instructions,
            **kwargs,
        }

        # Input audio transcription
        if self.enable_input_audio_transcription:
            session_config["audio"]["input"]["transcription"] = {
                "model": "whisper-1",
            }

        # Tools configuration
        if tools:
            session_config["tools"] = self._format_toolkit_schema(tools)

        return {
            "type": "session.update",
            "session": session_config,
        }

    def _format_toolkit_schema(
        self,
        schemas: list[dict[str, Any]],
    ) -> list[dict[str, Any]]:
        """Format the tools JSON schema into OpenAI realtime model format.

        Args:
            schemas (`list[dict[str, Any]]`):
                The tool schemas.

        Returns:
            `list[dict[str, Any]]`:
                The formatted tools for OpenAI realtime model.

        .. note::
            The OpenAI Realtime API uses a different tool format compared to
            the regular Chat Completions API. While the Chat API expects tools
            to be wrapped in ``{"type": "function", "function": {...}}``, the
            Realtime API expects a flattened structure where the function
            definition is directly at the top level with an added ``"type":
            "function"`` field.
        """
        return [{"type": "function", **tool["function"]} for tool in schemas]

    async def send(
        self,
        data: AudioBlock | TextBlock | ImageBlock | ToolResultBlock,
    ) -> None:
        """Send the data to the OpenAI realtime model for processing.

        Args:
            data (`AudioBlock | TextBlock | ImageBlock | ToolResultBlock`):
                The data to be sent to the OpenAI realtime model.
        """
        from websockets import State

        if not self._websocket or self._websocket.state != State.OPEN:
            raise RuntimeError(
                f"WebSocket is not connected for model {self.model_name}. "
                "Call the `connect` method first.",
            )

        # Type checking
        assert (
            isinstance(data, dict) and "type" in data
        ), "Data must be a dict with a 'type' field."

        # The source must be base64 for audio data
        data_type = data.get("type")

        if data_type not in self.support_input_modalities:
            logger.warning(
                "OpenAI Realtime API does not support %s data input. "
                "Supported modalities are: %s",
                data_type,
                ", ".join(self.support_input_modalities),
            )
            return

        # Process the data based on its type
        if data_type == "audio":
            to_send_message = await self._parse_audio_data(
                AudioBlock(
                    type="audio",
                    source=data.get("source"),
                ),
            )

        elif data_type == "text":
            to_send_message = await self._parse_text_data(
                TextBlock(
                    type="text",
                    text=data.get("text"),
                ),
            )

        elif data_type == "tool_result":
            to_send_message = await self._parse_tool_result_data(
                ToolResultBlock(
                    type="tool_result",
                    id=data.get("id"),
                    output=data.get("output"),
                    name=data.get("name"),
                ),
            )

        else:
            raise RuntimeError(f"Unsupported data type {data_type}")

        await self._websocket.send(to_send_message)

    async def parse_api_message(
        self,
        message: str,
    ) -> ModelEvents.EventBase | list[ModelEvents.EventBase] | None:
        """Parse the message received from the OpenAI realtime model API.

        Args:
            message (`str`):
                The message received from the OpenAI realtime model API.

        Returns:
            `ModelEvents.EventBase | list[ModelEvents.EventBase] | None`:
                The unified model event(s) in agentscope format.
        """
        try:
            data = json.loads(message)
        except json.decoder.JSONDecodeError:
            return None

        if not isinstance(data, dict):
            return None

        model_event = None
        match data.get("type", ""):
            # ================ Session related events ================
            case "session.created":
                model_event = ModelEvents.ModelSessionCreatedEvent(
                    session_id=data.get("session", {}).get("id", ""),
                )

            case "session.updated":
                # TODO: handle the session updated event
                pass

            # ================ Response related events ================
            case "response.created":
                self._response_id = data.get("response", {}).get("id", "")
                model_event = ModelEvents.ModelResponseCreatedEvent(
                    response_id=self._response_id,
                )

            case "response.done":
                response = data.get("response", {})
                response_id = response.get("id", self._response_id)
                usage = response.get("usage", {})
                model_event = ModelEvents.ModelResponseDoneEvent(
                    response_id=response_id,
                    input_tokens=usage.get("input_tokens", 0),
                    output_tokens=usage.get("output_tokens", 0),
                )
                # clear the response id
                self._response_id = ""

            case "response.output_audio.delta":
                audio_data = data.get("delta", "")
                if audio_data:
                    model_event = ModelEvents.ModelResponseAudioDeltaEvent(
                        response_id=self._response_id,
                        item_id=data.get("item_id", ""),
                        delta=audio_data,
                        format={
                            "type": "audio/pcm",
                            "rate": self.output_sample_rate,
                        },
                    )

            case "response.output_audio.done":
                model_event = ModelEvents.ModelResponseAudioDoneEvent(
                    response_id=self._response_id,
                    item_id=data.get("item_id", ""),
                )

            # ================ Transcription related events ================
            case "response.output_audio_transcript.delta":
                transcript_data = data.get("delta", "")
                if transcript_data:
                    model_event = (
                        ModelEvents.ModelResponseAudioTranscriptDeltaEvent(
                            response_id=self._response_id,
                            delta=transcript_data,
                            item_id=data.get("item_id", ""),
                        )
                    )

            case "response.output_audio_transcript.done":
                model_event = (
                    ModelEvents.ModelResponseAudioTranscriptDoneEvent(
                        response_id=self._response_id,
                        item_id=data.get("item_id", ""),
                    )
                )

            case "response.function_call_arguments.delta":
                arguments_delta = data.get("delta")
                call_id = data.get("call_id", "")
                if arguments_delta:
                    # Accumulate arguments
                    if call_id not in self._tool_args_accumulator:
                        self._tool_args_accumulator[call_id] = ""
                    self._tool_args_accumulator[call_id] += arguments_delta

                    # Return the accumulated arguments instead of just the
                    # delta
                    model_event = ModelEvents.ModelResponseToolUseDeltaEvent(
                        response_id=self._response_id,
                        item_id=data.get("item_id", ""),
                        tool_use=ToolUseBlock(
                            type="tool_use",
                            id=call_id,
                            name=data.get("name", ""),
                            input={},
                            raw_input=self._tool_args_accumulator[call_id],
                        ),
                    )
                    # TODO: This handles only one tool call at a time. For
                    #  parallel tool calls, we might need to reconsider the
                    #  event handling mechanism.

            case "response.function_call_arguments.done":
                call_id = data.get("call_id", "")
                current_input = self._tool_args_accumulator[call_id]
                model_event = ModelEvents.ModelResponseToolUseDoneEvent(
                    response_id=self._response_id,
                    item_id=data.get("item_id", ""),
                    tool_use=ToolUseBlock(
                        type="tool_use",
                        id=call_id,
                        name=data.get("name", ""),
                        input=_json_loads_with_repair(current_input),
                        raw_input=current_input,
                    ),
                )
                # Clear the accumulator for this call_id when done
                if call_id in self._tool_args_accumulator:
                    del self._tool_args_accumulator[call_id]

            case "conversation.item.input_audio_transcription.delta":
                delta = data.get("delta", "")
                if delta:
                    model_event = (
                        ModelEvents.ModelInputTranscriptionDeltaEvent(
                            item_id=data.get("item_id", ""),
                            delta=delta,
                        )
                    )

            case "conversation.item.input_audio_transcription.completed":
                transcript_data = data.get("transcript", "")
                if transcript_data:
                    model_event = ModelEvents.ModelInputTranscriptionDoneEvent(
                        transcript=transcript_data,
                        item_id=data.get("item_id", ""),
                    )

            # ================= VAD related events =================
            case "input_audio_buffer.speech_started":
                model_event = ModelEvents.ModelInputStartedEvent(
                    item_id=data.get("item_id", ""),
                    audio_start_ms=data.get("audio_start_ms", 0),
                )

            case "input_audio_buffer.speech_stopped":
                model_event = ModelEvents.ModelInputDoneEvent(
                    item_id=data.get("item_id", ""),
                    audio_end_ms=data.get("audio_end_ms", 0),
                )

            # ================= Error events =================
            case "error":
                error = data.get("error", {})
                model_event = ModelEvents.ModelErrorEvent(
                    error_type=error.get("type", "unknown"),
                    code=error.get("code", "unknown"),
                    message=error.get("message", "An unknown error occurred."),
                )

            # ================= Unknown events =================
            case _:
                logger.debug(
                    "Unknown OpenAI realtime model event type: %s",
                    data.get("type", None),
                )

        return model_event

    async def _parse_audio_data(self, block: AudioBlock) -> str:
        """Parse the audio data block to the format required by the OpenAI
        realtime model API.

        Args:
            block (`AudioBlock`):
                The audio data block.

        Returns:
            `str`: The parsed message to be sent to the OpenAI realtime
            model API.
        """
        if block["source"]["type"] == "base64":
            audio_data = block["source"]["data"]

        elif block["source"]["type"] == "url":
            audio_data = _get_bytes_from_web_url(block["source"]["url"])

        else:
            raise ValueError(
                f"Unsupported audio source type: {block['source']['type']}",
            )

        return json.dumps(
            {
                "type": "input_audio_buffer.append",
                "audio": audio_data,
            },
        )

    async def _parse_text_data(self, block: TextBlock) -> str:
        """Parse the text data block to the format required by the OpenAI
        realtime model API.

        Args:
            block (`TextBlock`):
                The text data block.

        Returns:
            `str`: The parsed message to be sent to the OpenAI realtime
            model API.
        """
        text = block.get("text", "")

        return json.dumps(
            {
                "type": "conversation.item.create",
                "item": {
                    "type": "message",
                    "role": "user",
                    "content": [
                        {
                            "type": "input_text",
                            "text": text,
                        },
                    ],
                },
            },
        )

    async def _parse_tool_result_data(self, block: ToolResultBlock) -> str:
        """Parse the tool result data block to the format required by the
        OpenAI realtime model API.

        Args:
            block (`ToolResultBlock`):
                The tool result data block.

        Returns:
            `str`: The parsed message to be sent to the OpenAI realtime
            model API.
        """
        return json.dumps(
            {
                "type": "conversation.item.create",
                "item": {
                    "type": "function_call_output",
                    "call_id": block.get("id"),
                    "output": block.get("output"),
                },
            },
        )



# -*- coding: utf-8 -*-
"""The client events for web-to-backend communication."""
from enum import Enum
from typing import List

from pydantic import BaseModel

from ._utils import AudioFormat
from ...message import TextBlock, AudioBlock, ImageBlock, VideoBlock


class ClientEventType(str, Enum):
    """Types of client events for web-to-backend communication."""

    # ============== Session control ================
    CLIENT_SESSION_CREATE = "client_session_create"
    """The user creates a new session in the frontend."""

    CLIENT_SESSION_END = "client_session_end"
    """The user ends the current session in the frontend."""

    # ============== Response control ================
    CLIENT_RESPONSE_CREATE = "client_response_create"
    """The user requests the agent to generate a response immediately."""

    CLIENT_RESPONSE_CANCEL = "client_response_cancel"
    """The user interrupts the agent's current response generation."""

    CLIENT_IMAGE_APPEND = "client_image_append"
    """The user appends an image input to the current session."""

    CLIENT_TEXT_APPEND = "client_text_append"
    """The user appends a text input to the current session."""

    CLIENT_AUDIO_APPEND = "client_audio_append"
    """The user appends an audio input to the current session."""

    CLIENT_AUDIO_COMMIT = "client_audio_commit"
    """The user commits the audio input to signal end of input."""

    CLIENT_TOOL_RESULT = "client_tool_result"
    """The tool result executed in the frontend is sent back to the backend."""


class ClientEvents:
    """Realtime client events."""

    class EventBase(BaseModel):
        """The base class for all client events, used to unify the type
        hinting."""

    class ClientSessionCreateEvent(EventBase):
        """Session create event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_SESSION_CREATE
        """The event type."""

        config: dict
        """The session config."""

    class ClientSessionEndEvent(EventBase):
        """Session end event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_SESSION_END
        """The event type."""

        session_id: str
        """The session ID."""

    class ClientResponseCreateEvent(EventBase):
        """Response create event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_RESPONSE_CREATE
        """The event type."""

        session_id: str
        """The session ID."""

    class ClientResponseCancelEvent(EventBase):
        """Response cancel event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_RESPONSE_CANCEL
        """The event type."""

        session_id: str
        """The session ID."""

    class ClientImageAppendEvent(EventBase):
        """Image append event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_IMAGE_APPEND
        """The event type."""

        session_id: str
        """The session ID."""

        image: str
        """The image data, encoded as base64 string."""

        format: dict
        """The image format information."""

    class ClientTextAppendEvent(EventBase):
        """Text append event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_TEXT_APPEND
        """The event type."""

        session_id: str
        """The session ID."""

        text: str
        """The text data."""

    class ClientAudioAppendEvent(EventBase):
        """Audio append event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_AUDIO_APPEND
        """The event type."""

        session_id: str
        """The session ID."""

        audio: str
        """The audio data, encoded as base64 string."""

        format: AudioFormat
        """The audio format information."""

    class ClientAudioCommitEvent(EventBase):
        """Audio commit event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_AUDIO_COMMIT
        """The event type."""

        session_id: str
        """The session ID."""

    class ClientToolResultEvent(EventBase):
        """Tool result event in the frontend"""

        type: ClientEventType = ClientEventType.CLIENT_TOOL_RESULT
        """The event type."""

        session_id: str
        """The session ID."""

        id: str
        """The tool call ID."""

        name: str
        """The tool name."""

        output: str | List[TextBlock | ImageBlock | AudioBlock | VideoBlock]
        """The tool result."""

    MAPPING = {
        ClientEventType.CLIENT_SESSION_CREATE: ClientSessionCreateEvent,
        ClientEventType.CLIENT_SESSION_END: ClientSessionEndEvent,
        ClientEventType.CLIENT_RESPONSE_CREATE: ClientResponseCreateEvent,
        ClientEventType.CLIENT_RESPONSE_CANCEL: ClientResponseCancelEvent,
        ClientEventType.CLIENT_IMAGE_APPEND: ClientImageAppendEvent,
        ClientEventType.CLIENT_TEXT_APPEND: ClientTextAppendEvent,
        ClientEventType.CLIENT_AUDIO_APPEND: ClientAudioAppendEvent,
        ClientEventType.CLIENT_AUDIO_COMMIT: ClientAudioCommitEvent,
        ClientEventType.CLIENT_TOOL_RESULT: ClientToolResultEvent,
    }

    @classmethod
    def from_json(cls, json_data: dict) -> EventBase:
        """Parse the client event from JSON data and return the corresponding
        event instance.

        Args:
            json_data (`dict`):
                The JSON data, which must contain the "type" field.

        Raises:
            `ValueError`:
                If the event type is unknown.

        Returns:
            `ClientEvents.EventBase`:
                The corresponding client event instance.
        """
        if not isinstance(json_data, dict) or "type" not in json_data:
            raise ValueError(
                f"Invalid JSON data for ClientEvent: {json_data}",
            )

        event_type = json_data["type"]

        if event_type not in cls.MAPPING:
            raise ValueError(f"Unknown ClientEvent type: {event_type}")

        # Obtain the event class from the mapping
        event_class = cls.MAPPING[event_type]
        return event_class(**json_data)



# -*- coding: utf-8 -*-
"""The working memory module in AgentScope, which provides various memory
storage implementations. In AgentScope, such module is responsible for
storing and managing the short-term memory with specific marks."""

from ._base import MemoryBase
from ._in_memory_memory import InMemoryMemory
from ._redis_memory import RedisMemory
from ._sqlalchemy_memory import AsyncSQLAlchemyMemory

__all__ = [
    "MemoryBase",
    "InMemoryMemory",
    "RedisMemory",
    "AsyncSQLAlchemyMemory",
]
